import os

from mediafile import MediaFile

from beets import library
from beets.test import _common
from beets.test.helper import BeetsTestCase, IOMixin
from beets.ui.commands.update import update_items
from beets.util import MoveOperation, remove, syspath


class UpdateTest(IOMixin, BeetsTestCase):
    def setUp(self):
        super().setUp()

        # Copy a file into the library.
        item_path = _common.RSRC / "full.mp3"
        item_path_two = _common.RSRC / "full.flac"
        self.i = library.Item.from_path(item_path)
        self.i2 = library.Item.from_path(item_path_two)
        self.lib.add(self.i)
        self.lib.add(self.i2)
        self.i.move(operation=MoveOperation.COPY)
        self.i2.move(operation=MoveOperation.COPY)
        self.album = self.lib.add_album([self.i, self.i2])

        # Album art.
        artfile = self.temp_path / "testart.jpg"
        artfile.touch()
        self.album.set_art(artfile)
        self.album.store()
        artfile.unlink()

    def _update(
        self,
        query=(),
        album=False,
        move=False,
        reset_mtime=True,
        fields=None,
        exclude_fields=None,
    ):
        self.io.addinput("y")
        if reset_mtime:
            self.i.mtime = 0
            self.i.store()
        update_items(
            self.lib,
            query,
            album,
            move,
            False,
            fields=fields,
            exclude_fields=exclude_fields,
        )

    def test_delete_removes_item(self):
        assert list(self.lib.items())
        remove(self.i.path)
        remove(self.i2.path)
        self._update()
        assert not list(self.lib.items())

    def test_delete_removes_album(self):
        assert self.lib.albums()
        remove(self.i.path)
        remove(self.i2.path)
        self._update()
        assert not self.lib.albums()

    def test_delete_removes_album_art(self):
        art_filepath = self.album.art_filepath
        assert art_filepath.exists()
        remove(self.i.path)
        remove(self.i2.path)
        self._update()
        assert not art_filepath.exists()

    def test_modified_metadata_detected(self):
        mf = MediaFile(syspath(self.i.path))
        mf.title = "differentTitle"
        mf.save()
        self._update()
        item = self.lib.items().get()
        assert item.title == "differentTitle"

    def test_modified_metadata_moved(self):
        mf = MediaFile(syspath(self.i.path))
        mf.title = "differentTitle"
        mf.save()
        self._update(move=True)
        item = self.lib.items().get()
        assert b"differentTitle" in item.path

    def test_modified_metadata_not_moved(self):
        mf = MediaFile(syspath(self.i.path))
        mf.title = "differentTitle"
        mf.save()
        self._update(move=False)
        item = self.lib.items().get()
        assert b"differentTitle" not in item.path

    def test_selective_modified_metadata_moved(self):
        mf = MediaFile(syspath(self.i.path))
        mf.title = "differentTitle"
        mf.genres = ["differentGenre"]
        mf.save()
        self._update(move=True, fields=["title"])
        item = self.lib.items().get()
        assert b"differentTitle" in item.path
        assert item.genres != ["differentGenre"]

    def test_selective_modified_metadata_not_moved(self):
        mf = MediaFile(syspath(self.i.path))
        mf.title = "differentTitle"
        mf.genres = ["differentGenre"]
        mf.save()
        self._update(move=False, fields=["title"])
        item = self.lib.items().get()
        assert b"differentTitle" not in item.path
        assert item.genres != ["differentGenre"]

    def test_modified_album_metadata_moved(self):
        mf = MediaFile(syspath(self.i.path))
        mf.album = "differentAlbum"
        mf.save()
        self._update(move=True)
        item = self.lib.items().get()
        assert b"differentAlbum" in item.path

    def test_modified_album_metadata_art_moved(self):
        artpath = self.album.artpath
        mf = MediaFile(syspath(self.i.path))
        mf.album = "differentAlbum"
        mf.save()
        self._update(move=True)
        album = self.lib.albums()[0]
        assert artpath != album.artpath
        assert album.artpath is not None

    def test_selective_modified_album_metadata_moved(self):
        mf = MediaFile(syspath(self.i.path))
        mf.album = "differentAlbum"
        mf.genres = ["differentGenre"]
        mf.save()
        self._update(move=True, fields=["album"])
        item = self.lib.items().get()
        assert b"differentAlbum" in item.path
        assert item.genres != ["differentGenre"]

    def test_selective_modified_album_metadata_not_moved(self):
        mf = MediaFile(syspath(self.i.path))
        mf.album = "differentAlbum"
        mf.genres = ["differentGenre"]
        mf.save()
        self._update(move=True, fields=["genres"])
        item = self.lib.items().get()
        assert b"differentAlbum" not in item.path
        assert item.genres == ["differentGenre"]

    def test_mtime_match_skips_update(self):
        mf = MediaFile(syspath(self.i.path))
        mf.title = "differentTitle"
        mf.save()

        # Make in-memory mtime match on-disk mtime.
        self.i.mtime = os.path.getmtime(syspath(self.i.path))
        self.i.store()

        self._update(reset_mtime=False)
        item = self.lib.items().get()
        assert item.title == "full"

    def test_multivalued_albumtype_roundtrip(self):
        # https://github.com/beetbox/beets/issues/4528

        # albumtypes is empty for our test fixtures, so populate it first
        album = self.album
        correct_albumtypes = ["album", "live"]

        # Setting albumtypes does not set albumtype, currently.
        # Using x[0] mirrors https://github.com/beetbox/mediafile/blob/057432ad53b3b84385e5582f69f44dc00d0a725d/mediafile.py#L1928  # noqa: E501
        correct_albumtype = correct_albumtypes[0]

        album.albumtype = correct_albumtype
        album.albumtypes = correct_albumtypes
        album.try_sync(write=True, move=False)

        album.load()
        assert album.albumtype == correct_albumtype
        assert album.albumtypes == correct_albumtypes

        self._update()

        album.load()
        assert album.albumtype == correct_albumtype
        assert album.albumtypes == correct_albumtypes

    def test_modified_metadata_excluded(self):
        mf = MediaFile(syspath(self.i.path))
        mf.lyrics = "new lyrics"
        mf.save()
        self._update(exclude_fields=["lyrics"])
        item = self.lib.items().get()
        assert item.lyrics != "new lyrics"



from __future__ import annotations

from typing import Generic, Literal, TypedDict, TypeVar

from typing_extensions import NotRequired


class ResourceIdentifier(TypedDict):
    id: str
    type: str


class RelationshipLinks(TypedDict):
    self: NotRequired[str]
    next: NotRequired[str]


class RelationshipData(TypedDict):
    data: list[ResourceIdentifier]
    links: RelationshipLinks


class SingleRelationshipData(TypedDict):
    data: ResourceIdentifier
    links: NotRequired[RelationshipLinks]


class ExternalLink(TypedDict):
    href: str
    meta: str


class Copyright(TypedDict):
    text: str


class ArtistAttributes(TypedDict):
    name: str
    popularity: float  # 0.0 - 1.0, required
    handle: NotRequired[str]
    ownerType: NotRequired[Literal["LABEL", "USER", "MIXED"]]
    spotlighted: NotRequired[bool]


class MediaAttributes(TypedDict):
    """Describe media metadata exposed by the TIDAL API.

    Combines the core values needed to identify and present a media item with
    optional publishing, attribution, and sharing details when TIDAL provides
    them.

    While this type is not part of the original TIDAL json:api schema itself, we
    introduced it to simplify our type definitions and allow for easier reuse of
    shared attributes.
    """

    duration: str  # ISO 8601
    title: str
    explicit: bool
    mediaTags: list[str]
    popularity: float

    accessType: NotRequired[Literal["PUBLIC", "UNLISTED", "PRIVATE"]]
    copyright: NotRequired[Copyright]
    createdAt: NotRequired[str]  # ISO 8601 datetime
    externalLinks: NotRequired[list[ExternalLink]]
    version: NotRequired[str]


class AlbumAttributes(MediaAttributes):
    """Represent album-specific metadata returned by the TIDAL API.

    Extends shared media fields with release packaging details that describe how
    an album is published and how many items or volumes it contains.
    """

    # see "Albums_Attributes"
    # in https://tidal-music.github.io/tidal-api-reference/tidal-api-oas.json

    # Required
    albumType: Literal["ALBUM", "EP", "SINGLE"]
    barcodeId: str
    numberOfItems: int
    numberOfVolumes: int

    # Optional
    releaseDate: NotRequired[str]  # ISO date YYYY-MM-DD


class TrackAttributes(MediaAttributes):
    """Represent track-specific metadata returned by the TIDAL API.

    Adds the musical and catalog information needed to describe an individual
    recording, while allowing enrichment fields when TIDAL includes them.
    """

    # see "Tracks_Attributes"
    # in https://tidal-music.github.io/tidal-api-reference/tidal-api-oas.json

    # Required
    isrc: str
    key: Literal[
        "UNKNOWN",
        "C",
        "CSharp",
        "D",
        "Eb",
        "E",
        "F",
        "FSharp",
        "G",
        "Ab",
        "A",
        "Bb",
        "B",
    ]
    keyScale: Literal[
        "UNKNOWN",
        "MAJOR",
        "MINOR",
        "AEOLIAN",
        "BLUES",
        "DORIAN",
        "HARMONIC_MINOR",
        "LOCRIAN",
        "LYDIAN",
        "MIXOLYDIAN",
        "PENTATONIC_MAJOR",
        "PHRYGIAN",
        "MELODIC_MINOR",
        "PENTATONIC_MINOR",
    ]

    # Optional
    bpm: NotRequired[float]
    spotlighted: NotRequired[bool]
    toneTags: NotRequired[list[str]]


class SearchAttributes(TypedDict):
    didYouMean: NotRequired[str]
    trackingId: str


class TidalArtist(TypedDict):
    id: str
    type: Literal["artists"]
    attributes: ArtistAttributes


class TidalAlbum(TypedDict):
    id: str
    type: Literal["albums"]
    attributes: AlbumAttributes
    relationships: dict[str, RelationshipData]


class TidalTrack(TypedDict):
    id: str
    type: Literal["tracks"]
    attributes: TrackAttributes
    relationships: dict[str, RelationshipData]


class FileMeta(TypedDict):
    width: int
    height: int


class ArtworkFile(TypedDict):
    href: str
    meta: FileMeta


class ArtworkAttributes(TypedDict):
    mediaType: Literal["IMAGE"]
    files: list[ArtworkFile]
    visualMetadata: NotRequired[dict[str, str]]


class TidalArtwork(TypedDict):
    id: str
    type: Literal["artworks"]
    attributes: ArtworkAttributes


class TidalSearch(TypedDict):
    id: str
    type: Literal["searchResults"]
    attributes: SearchAttributes
    relationships: dict[str, RelationshipData]


T = TypeVar("T")


class Document(TypedDict, Generic[T]):
    data: T
    included: NotRequired[
        list[TidalArtist | TidalAlbum | TidalTrack | TidalArtwork]
    ]
    links: NotRequired[dict[str, str]]


AlbumDocument = Document[list[TidalAlbum]]
TrackDocument = Document[list[TidalTrack]]
SearchDocument = Document[TidalSearch]



"""DBCore is an abstract database package that forms the basis for beets'
Library.
"""

from .db import Database, Index, Model, Results
from .query import (
    AndQuery,
    FieldQuery,
    InvalidQueryError,
    MatchQuery,
    OrQuery,
    Query,
)
from .queryparse import (
    parse_sorted_query,
    query_from_strings,
    sort_from_strings,
)
from .types import Type

__all__ = [
    "AndQuery",
    "Database",
    "FieldQuery",
    "Index",
    "InvalidQueryError",
    "MatchQuery",
    "Model",
    "OrQuery",
    "Query",
    "Results",
    "Type",
    "parse_sorted_query",
    "query_from_strings",
    "sort_from_strings",
]



"""Tests for the 'hidden' utility."""

import ctypes
import errno
import subprocess
import sys
import tempfile
import unittest

from beets import util
from beets.util import bytestring_path, hidden


class HiddenFileTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_osx_hidden(self):
        if not sys.platform == "darwin":
            self.skipTest("sys.platform is not darwin")
            return

        with tempfile.NamedTemporaryFile(delete=False) as f:
            try:
                command = ["chflags", "hidden", f.name]
                subprocess.Popen(command).wait()
            except OSError as e:
                if e.errno == errno.ENOENT:
                    self.skipTest("unable to find chflags")
                else:
                    raise e

            assert hidden.is_hidden(bytestring_path(f.name))

    def test_windows_hidden(self):
        if not sys.platform == "win32":
            self.skipTest("sys.platform is not windows")
            return

        # FILE_ATTRIBUTE_HIDDEN = 2 (0x2) from GetFileAttributes documentation.
        hidden_mask = 2

        with tempfile.NamedTemporaryFile() as f:
            # Hide the file using
            success = ctypes.windll.kernel32.SetFileAttributesW(
                f.name, hidden_mask
            )

            if not success:
                self.skipTest("unable to set file attributes")

            assert hidden.is_hidden(f.name)

    def test_other_hidden(self):
        if sys.platform == "darwin" or sys.platform == "win32":
            self.skipTest("sys.platform is known")
            return

        with tempfile.NamedTemporaryFile(prefix=".tmp") as f:
            fn = util.bytestring_path(f.name)
            assert hidden.is_hidden(fn)



"""Utility functions for beets UI commands."""

from beets.exceptions import UserError


def do_query(lib, query, album, also_items=True):
    """For commands that operate on matched items, performs a query
    and returns a list of matching items and a list of matching
    albums. (The latter is only nonempty when album is True.) Raises
    a UserError if no items match. also_items controls whether, when
    fetching albums, the associated items should be fetched also.
    """
    if album:
        albums = list(lib.albums(query))
        items = []
        if also_items:
            for al in albums:
                items += al.items()

    else:
        albums = []
        items = list(lib.items(query))

    if album and not albums:
        raise UserError("No matching albums found.")
    if not album and not items:
        raise UserError("No matching items found.")

    return items, albums



"""Tests for non-query database functions of Item."""

from __future__ import annotations

import os
import os.path
import re
import shutil
import stat
import unittest
from pathlib import Path
from unittest.mock import patch

import pytest
from mediafile import MediaFile, UnreadableFileError

import beets.dbcore.query
import beets.library
from beets import config, plugins, util
from beets.library import Album
from beets.test import _common
from beets.test._common import item
from beets.test.helper import TestHelper
from beets.util import as_string, bytestring_path, normpath, syspath

# Shortcut to path normalization.
np = util.normpath


class PytestItemHelper(TestHelper):
    def get_first_item(self):
        """Retrieve first item from library."""
        return next(iter(self.lib.items()))

    @pytest.fixture
    def item(self):
        return _common.item()

    @pytest.fixture
    def item_in_db(self):
        return _common.item(self.lib)


class TestLoad(PytestItemHelper):
    def test_load_restores_data_from_db(self, item_in_db):
        original_title = item_in_db.title
        item_in_db.title = "something"
        item_in_db.load()
        assert original_title == item_in_db.title

    def test_load_clears_dirty_flags(self, item_in_db):
        item_in_db.artist = "something"
        assert "artist" in item_in_db._dirty
        item_in_db.load()
        assert "artist" not in item_in_db._dirty


class TestStore(PytestItemHelper):
    def test_store_changes_database_value(self, item_in_db):
        new_year = 1987
        item_in_db.year = new_year
        item_in_db.store()

        assert self.lib.get_item(item_in_db.id).year == new_year

    def test_store_only_writes_dirty_fields(self, item_in_db):
        new_year = 1987
        item_in_db._values_fixed["year"] = new_year  # change w/o dirtying
        item_in_db.store()

        assert self.lib.get_item(item_in_db.id).year != new_year

    def test_store_clears_dirty_flags(self, item_in_db):
        item_in_db.composers = ["tvp"]
        item_in_db.store()
        assert "composers" not in item_in_db._dirty

    def test_store_album_cascades_flex_deletes(self):
        album = Album(flex1="Flex-1")
        self.lib.add(album)
        item = _common.item()
        item.album_id = album.id
        item.flex1 = "Flex-1"
        self.lib.add(item)
        del album.flex1
        album.store()
        assert "flex1" not in album
        assert "flex1" not in album.items()[0]

    def test_store_does_not_propagate_artpath_to_items(self):
        item = _common.item()
        self.lib.add(item)
        album = self.lib.add_album([item])
        assert "artpath" not in Album.item_keys
        album.artpath = b"/abs/path/to/cover.jpg"
        album.store()
        stored = self.lib.get_item(item.id)
        assert not stored.get("artpath", with_album=False)


class TestAdd(PytestItemHelper):
    def test_item_add_inserts_row(self, item):
        self.lib.add(item)
        new_grouping = (
            self.lib._connection()
            .execute(
                "select grouping from items where composers = ?",
                (item._type("composers").to_sql(item.composers),),
            )
            .fetchone()["grouping"]
        )
        assert new_grouping == item.grouping

    def test_library_add_path_inserts_row(self):
        item = beets.library.Item.from_path(_common.RSRC / "full.mp3")
        self.lib.add(item)
        new_grouping = (
            self.lib._connection()
            .execute(
                "select grouping from items where composers = ?",
                (item._type("composers").to_sql(item.composers),),
            )
            .fetchone()["grouping"]
        )
        assert new_grouping == item.grouping

    def test_library_add_one_database_change_event(
        self, item, caplog: pytest.LogCaptureFixture
    ):
        """Test library.add emits only one database_change event."""

        item.path = beets.util.normpath(self.temp_path / "a" / "b.mp3")
        item.album = "a"
        item.title = "b"

        with caplog.at_level("DEBUG", logger="beets"):
            self.lib.add(item)

        assert caplog.text.count("Sending event: database_change") == 1


class TestRemove(PytestItemHelper):
    def test_remove_deletes_from_db(self, item_in_db):
        item_in_db.remove()
        c = self.lib._connection().execute("select * from items")
        assert c.fetchone() is None


class TestGetSet(PytestItemHelper):
    def test_set_changes_value(self, item):
        item.bpm = 4915
        assert item.bpm == 4915

    def test_set_sets_dirty_flag(self, item):
        item.comp = not item.comp
        assert "comp" in item._dirty

    def test_set_does_not_dirty_if_value_unchanged(self, item):
        item.title = item.title
        assert "title" not in item._dirty

    def test_invalid_field_raises_attributeerror(self, item):
        with pytest.raises(AttributeError):
            item.xyzzy

    def test_album_fallback(self, item_in_db):
        # integration test of item-album fallback
        album = self.lib.add_album([item_in_db])
        album["flex"] = "foo"
        album.store()

        assert "flex" in item_in_db
        assert "flex" not in item_in_db.keys(with_album=False)
        assert item_in_db["flex"] == "foo"
        assert item_in_db.get("flex") == "foo"
        assert item_in_db.get("flex", with_album=False) is None
        assert item_in_db.get("flexx") is None


class TestDestination(PytestItemHelper):
    """Confirm tests handle temporary directory path containing '.'"""

    def create_temp_dir(self, **kwargs):
        kwargs["prefix"] = "."
        return super().create_temp_dir(**kwargs)

    def test_directory_works_with_trailing_slash(self, item_in_db):
        self.lib.directory = b"one/"
        self.lib.path_formats = [("default", "two")]
        assert item_in_db.destination() == np("one/two")

    def test_directory_works_without_trailing_slash(self, item_in_db):
        self.lib.directory = b"one"
        self.lib.path_formats = [("default", "two")]
        assert item_in_db.destination() == np("one/two")

    def test_destination_substitutes_metadata_values(self, item_in_db):
        self.lib.directory = b"base"
        self.lib.path_formats = [("default", "$album/$artist $title")]
        item_in_db.title = "three"
        item_in_db.artist = "two"
        item_in_db.album = "one"
        assert item_in_db.destination() == np("base/one/two three")

    def test_destination_preserves_extension(self, item_in_db):
        self.lib.directory = b"base"
        self.lib.path_formats = [("default", "$title")]
        item_in_db.path = "hey.audioformat"
        assert item_in_db.destination() == np("base/the title.audioformat")

    def test_lower_case_extension(self, item_in_db):
        self.lib.directory = b"base"
        self.lib.path_formats = [("default", "$title")]
        item_in_db.path = "hey.MP3"
        assert item_in_db.destination() == np("base/the title.mp3")

    def test_destination_pads_some_indices(self, item_in_db):
        self.lib.directory = b"base"
        self.lib.path_formats = [
            ("default", "$track $tracktotal $disc $disctotal $bpm")
        ]
        item_in_db.track = 1
        item_in_db.tracktotal = 2
        item_in_db.disc = 3
        item_in_db.disctotal = 4
        item_in_db.bpm = 5
        assert item_in_db.destination() == np("base/01 02 03 04 5")

    def test_destination_pads_date_values(self, item_in_db):
        self.lib.directory = b"base"
        self.lib.path_formats = [("default", "$year-$month-$day")]
        item_in_db.year = 1
        item_in_db.month = 2
        item_in_db.day = 3
        assert item_in_db.destination() == np("base/0001-02-03")

    def test_destination_escapes_slashes(self, item_in_db):
        self.lib.path_formats = [("default", "$artist/$album/$track $title")]
        item_in_db.album = "one/two"
        dest = item_in_db.destination()
        assert b"one" in dest
        assert b"two" in dest
        assert b"one/two" not in dest

    def test_destination_escapes_leading_dot(self, item_in_db):
        self.lib.path_formats = [("default", "$artist/$album/$track $title")]
        item_in_db.album = ".something"
        dest = item_in_db.destination()
        assert b"something" in dest
        assert b"/.something" not in dest

    def test_destination_preserves_legitimate_slashes(self, item_in_db):
        self.lib.path_formats = [("default", "$artist/$album/$track $title")]
        item_in_db.artist = "one"
        item_in_db.album = "two"
        dest = item_in_db.destination()
        assert os.path.join(b"one", b"two") in dest

    def test_destination_long_names_truncated(self, item_in_db):
        item_in_db.title = "X" * 300
        item_in_db.artist = "Y" * 300
        for c in item_in_db.destination().split(util.PATH_SEP):
            assert len(c) <= 255

    def test_destination_long_names_keep_extension(self, item_in_db):
        item_in_db.title = "X" * 300
        item_in_db.path = b"something.extn"
        dest = item_in_db.destination()
        assert dest[-5:] == b".extn"

    def test_distination_windows_removes_both_separators(self, item_in_db):
        item_in_db.title = "one \\ two / three.mp3"
        with _common.platform_windows():
            p = item_in_db.destination()
        assert b"one \\ two" not in p
        assert b"one / two" not in p
        assert b"two \\ three" not in p
        assert b"two / three" not in p

    def test_path_with_format(self, item_in_db):
        self.lib.path_formats = [("default", "$artist/$album ($format)")]
        p = item_in_db.destination()
        assert b"(FLAC)" in p

    def test_heterogeneous_album_gets_single_directory(self):
        i1, i2 = _common.item(self.lib), _common.item(self.lib)
        self.lib.add_album([i1, i2])
        i1.year, i2.year = 2009, 2010
        self.lib.path_formats = [("default", "$album ($year)/$track $title")]
        dest1, dest2 = i1.destination(), i2.destination()
        assert os.path.dirname(dest1) == os.path.dirname(dest2)

    def test_default_path_for_non_compilations(self, item_in_db):
        item_in_db.comp = False
        self.lib.add_album([item_in_db])
        self.lib.directory = b"one"
        self.lib.path_formats = [("default", "two"), ("comp:true", "three")]
        assert item_in_db.destination() == np("one/two")

    def test_singleton_path(self, item_in_db):
        self.lib.directory = b"one"
        self.lib.path_formats = [
            ("default", "two"),
            ("singleton:true", "four"),
            ("comp:true", "three"),
        ]
        assert item_in_db.destination() == np("one/four")

    def test_comp_before_singleton_path(self, item_in_db):
        item_in_db.comp = True
        self.lib.directory = b"one"
        self.lib.path_formats = [
            ("default", "two"),
            ("comp:true", "three"),
            ("singleton:true", "four"),
        ]
        assert item_in_db.destination() == np("one/three")

    def test_comp_path(self, item_in_db):
        item_in_db.comp = True
        self.lib.add_album([item_in_db])
        self.lib.directory = b"one"
        self.lib.path_formats = [("default", "two"), ("comp:true", "three")]
        assert item_in_db.destination() == np("one/three")

    def test_multi_value_string_query_path(self, item_in_db):
        item_in_db.genres = ["Classical"]
        self.lib.directory = b"one"
        self.lib.path_formats = [
            ("default", "two"),
            ("genres:=~Classical", "three"),
        ]
        assert item_in_db.destination() == np("one/three")

    def test_multi_value_match_query_path(self, item_in_db):
        item_in_db.genres = ["Classical"]
        self.lib.directory = b"one"
        self.lib.path_formats = [
            ("default", "two"),
            ("genres:=Classical", "three"),
        ]
        assert item_in_db.destination() == np("one/three")

    def test_multi_value_string_query_path_no_substring_match(self, item_in_db):
        item_in_db.genres = ["Neoclassical"]
        self.lib.directory = b"one"
        self.lib.path_formats = [
            ("default", "two"),
            ("genres:=~Classical", "three"),
        ]
        assert item_in_db.destination() == np("one/two")

    def test_albumtype_query_path(self, item_in_db):
        item_in_db.comp = True
        self.lib.add_album([item_in_db])
        item_in_db.albumtype = "sometype"
        self.lib.directory = b"one"
        self.lib.path_formats = [
            ("default", "two"),
            ("albumtype:sometype", "four"),
            ("comp:true", "three"),
        ]
        assert item_in_db.destination() == np("one/four")

    def test_albumtype_path_fallback_to_comp(self, item_in_db):
        item_in_db.comp = True
        self.lib.add_album([item_in_db])
        item_in_db.albumtype = "sometype"
        self.lib.directory = b"one"
        self.lib.path_formats = [
            ("default", "two"),
            ("albumtype:anothertype", "four"),
            ("comp:true", "three"),
        ]
        assert item_in_db.destination() == np("one/three")

    def test_get_formatted_does_not_replace_separators(self, item_in_db):
        with _common.platform_posix():
            name = os.path.join("a", "b")
            item_in_db.title = name
            newname = item_in_db.formatted().get("title")
        assert name == newname

    def test_get_formatted_pads_with_zero(self, item_in_db):
        with _common.platform_posix():
            item_in_db.track = 1
            name = item_in_db.formatted().get("track")
        assert name.startswith("0")

    def test_get_formatted_uses_kbps_bitrate(self, item_in_db):
        with _common.platform_posix():
            item_in_db.bitrate = 12345
            val = item_in_db.formatted().get("bitrate")
        assert val == "12kbps"

    def test_get_formatted_uses_khz_samplerate(self, item_in_db):
        with _common.platform_posix():
            item_in_db.samplerate = 12345
            val = item_in_db.formatted().get("samplerate")
        assert val == "12kHz"

    def test_get_formatted_datetime(self, item_in_db):
        with _common.platform_posix():
            item_in_db.added = 1368302461.210265
            val = item_in_db.formatted().get("added")
        assert val.startswith("2013")

    def test_get_formatted_none(self, item_in_db):
        with _common.platform_posix():
            item_in_db.some_other_field = None
            val = item_in_db.formatted().get("some_other_field")
        assert val == ""

    def test_artist_falls_back_to_albumartist(self, item_in_db):
        item_in_db.artist = ""
        item_in_db.albumartist = "something"
        self.lib.path_formats = [("default", "$artist")]
        p = item_in_db.destination()
        assert p.rsplit(util.PATH_SEP, 1)[1] == b"something"

    def test_albumartist_falls_back_to_artist(self, item_in_db):
        item_in_db.artist = "trackartist"
        item_in_db.albumartist = ""
        self.lib.path_formats = [("default", "$albumartist")]
        p = item_in_db.destination()
        assert p.rsplit(util.PATH_SEP, 1)[1] == b"trackartist"

    def test_artist_overrides_albumartist(self, item_in_db):
        item_in_db.artist = "theartist"
        item_in_db.albumartist = "something"
        self.lib.path_formats = [("default", "$artist")]
        p = item_in_db.destination()
        assert p.rsplit(util.PATH_SEP, 1)[1] == b"theartist"

    def test_albumartist_overrides_artist(self, item_in_db):
        item_in_db.artist = "theartist"
        item_in_db.albumartist = "something"
        self.lib.path_formats = [("default", "$albumartist")]
        p = item_in_db.destination()
        assert p.rsplit(util.PATH_SEP, 1)[1] == b"something"

    def test_unicode_extension_in_fragment(self, item_in_db):
        self.lib.path_formats = [("default", "foo")]
        item_in_db.path = util.bytestring_path("bar.caf\xe9")
        with patch("sys.platform", "linux"):
            dest = item_in_db.destination(relative_to_libdir=True)
        assert as_string(dest) == "foo.caf\xe9"

    def test_asciify_character_expanding_to_slash(self, item_in_db):
        config["asciify_paths"] = True
        self.lib.directory = b"lib"
        self.lib.path_formats = [("default", "$title")]
        item_in_db.title = "ab\xa2\xbdd"
        assert item_in_db.destination() == np("lib/abC_ 1_2d")

    def test_destination_with_replacements(self, item_in_db):
        self.lib.directory = b"base"
        self.lib.replacements = [(re.compile(r"a"), "e")]
        self.lib.path_formats = [("default", "$album/$title")]
        item_in_db.title = "foo"
        item_in_db.album = "bar"
        assert item_in_db.destination() == np("base/ber/foo")

    @unittest.skip("unimplemented: #359")
    def test_destination_with_empty_component(self, item_in_db):
        self.lib.directory = b"base"
        self.lib.replacements = [(re.compile(r"^$"), "_")]
        self.lib.path_formats = [("default", "$album/$artist/$title")]
        item_in_db.title = "three"
        item_in_db.artist = ""
        item_in_db.albumartist = ""
        item_in_db.album = "one"
        assert item_in_db.destination() == np("base/one/_/three")

    @unittest.skip("unimplemented: #359")
    def test_destination_with_empty_final_component(self, item_in_db):
        self.lib.directory = b"base"
        self.lib.replacements = [(re.compile(r"^$"), "_")]
        self.lib.path_formats = [("default", "$album/$title")]
        item_in_db.title = ""
        item_in_db.album = "one"
        item_in_db.path = "foo.mp3"
        assert item_in_db.destination() == np("base/one/_.mp3")

    def test_album_field_query(self, item_in_db):
        self.lib.directory = b"one"
        self.lib.path_formats = [("default", "two"), ("flex:foo", "three")]
        album = self.lib.add_album([item_in_db])
        assert item_in_db.destination() == np("one/two")
        album["flex"] = "foo"
        album.store()
        assert item_in_db.destination() == np("one/three")

    def test_album_field_in_template(self, item_in_db):
        self.lib.directory = b"one"
        self.lib.path_formats = [("default", "$flex/two")]
        album = self.lib.add_album([item_in_db])
        album["flex"] = "foo"
        album.store()
        assert item_in_db.destination() == np("one/foo/two")


class TestItemFormattedMapping(PytestItemHelper):
    def test_formatted_item_value(self, item_in_db):
        formatted = item_in_db.formatted()
        assert formatted["artist"] == "the artist"

    def test_get_unset_field(self, item_in_db):
        formatted = item_in_db.formatted()
        with pytest.raises(KeyError):
            formatted["other_field"]

    def test_get_method_with_default(self, item_in_db):
        formatted = item_in_db.formatted()
        assert formatted.get("other_field") == ""

    def test_get_method_with_specified_default(self, item_in_db):
        formatted = item_in_db.formatted()
        assert formatted.get("other_field", "default") == "default"

    def test_item_precedence(self, item_in_db):
        album = self.lib.add_album([item_in_db])
        album["artist"] = "foo"
        album.store()
        assert "foo" != item_in_db.formatted().get("artist")

    def test_album_flex_field(self, item_in_db):
        album = self.lib.add_album([item_in_db])
        album["flex"] = "foo"
        album.store()
        assert "foo" == item_in_db.formatted().get("flex")

    def test_album_field_overrides_item_field_for_path(self, item_in_db):
        # Make the album inconsistent with the item.
        album = self.lib.add_album([item_in_db])
        album.album = "foo"
        album.store()
        item_in_db.album = "bar"
        item_in_db.store()

        # Ensure the album takes precedence.
        formatted = item_in_db.formatted(for_path=True)
        assert formatted["album"] == "foo"

    def test_artist_falls_back_to_albumartist(self, item_in_db):
        item_in_db.artist = ""
        formatted = item_in_db.formatted()
        assert formatted["artist"] == "the album artist"

    def test_albumartist_falls_back_to_artist(self, item_in_db):
        item_in_db.albumartist = ""
        formatted = item_in_db.formatted()
        assert formatted["albumartist"] == "the artist"

    def test_both_artist_and_albumartist_empty(self, item_in_db):
        item_in_db.artist = ""
        item_in_db.albumartist = ""
        formatted = item_in_db.formatted()
        assert formatted["albumartist"] == ""


class PathFormattingMixin:
    """Utilities for testing path formatting."""

    lib: beets.library.Library

    def _setf(self, fmt):
        self.lib.path_formats.insert(0, ("default", fmt))

    def _assert_dest(self, dest, item):
        # Handle paths on Windows.
        if os.path.sep != "/":
            dest = dest.replace(b"/", os.path.sep.encode())

            # Paths are normalized based on the CWD.
            dest = normpath(dest)

        actual = item.destination()

        assert actual == dest


class TestDestinationFunction(TestHelper, PathFormattingMixin):
    @pytest.fixture(autouse=True)
    def item(self, setup):
        self.lib.directory = b"/base"
        self.lib.path_formats = [("default", "path")]
        return item(self.lib)

    def test_upper_case_literal(self, item):
        self._setf("%upper{foo}")
        self._assert_dest(b"/base/FOO", item)

    def test_upper_case_variable(self, item):
        self._setf("%upper{$title}")
        self._assert_dest(b"/base/THE TITLE", item)

    def test_capitalize_variable(self, item):
        self._setf("%capitalize{$title}")
        self._assert_dest(b"/base/The title", item)

    def test_title_case_variable(self, item):
        self._setf("%title{$title}")
        self._assert_dest(b"/base/The Title", item)

    def test_title_case_variable_aphostrophe(self, item):
        self._setf("%title{I can't}")
        self._assert_dest(b"/base/I Can't", item)

    def test_asciify_variable(self, item):
        self._setf("%asciify{ab\xa2\xbdd}")
        self._assert_dest(b"/base/abC_ 1_2d", item)

    def test_left_variable(self, item):
        self._setf("%left{$title, 3}")
        self._assert_dest(b"/base/the", item)

    def test_right_variable(self, item):
        self._setf("%right{$title,3}")
        self._assert_dest(b"/base/tle", item)

    def test_if_false(self, item):
        self._setf("x%if{,foo}")
        self._assert_dest(b"/base/x", item)

    def test_if_false_value(self, item):
        self._setf("x%if{false,foo}")
        self._assert_dest(b"/base/x", item)

    def test_if_true(self, item):
        self._setf("%if{bar,foo}")
        self._assert_dest(b"/base/foo", item)

    def test_if_else_false(self, item):
        self._setf("%if{,foo,baz}")
        self._assert_dest(b"/base/baz", item)

    def test_if_else_false_value(self, item):
        self._setf("%if{false,foo,baz}")
        self._assert_dest(b"/base/baz", item)

    def test_if_int_value(self, item):
        self._setf("%if{0,foo,baz}")
        self._assert_dest(b"/base/baz", item)

    def test_nonexistent_function(self, item):
        self._setf("%foo{bar}")
        self._assert_dest(b"/base/%foo{bar}", item)

    def test_if_def_field_return_self(self, item):
        item.bar = 3
        self._setf("%ifdef{bar}")
        self._assert_dest(b"/base/3", item)

    def test_if_def_field_not_defined(self, item):
        self._setf(" %ifdef{bar}/$artist")
        self._assert_dest(b"/base/the artist", item)

    def test_if_def_field_not_defined_2(self, item):
        self._setf("$artist/%ifdef{bar}")
        self._assert_dest(b"/base/the artist", item)

    def test_if_def_true(self, item):
        self._setf("%ifdef{artist,cool}")
        self._assert_dest(b"/base/cool", item)

    def test_if_def_true_complete(self, item):
        item.series = "Now"
        self._setf("%ifdef{series,$series Series,Albums}/$album")
        self._assert_dest(b"/base/Now Series/the album", item)

    def test_if_def_false_complete(self, item):
        self._setf("%ifdef{plays,$plays,not_played}")
        self._assert_dest(b"/base/not_played", item)

    def test_first(self, item):
        item.albumtypes = ["album", "compilation"]
        self._setf("%first{$albumtypes}")
        self._assert_dest(b"/base/album", item)

    def test_first_skip(self, item):
        item.albumtype = "album; ep; compilation"
        self._setf("%first{$albumtype,1,2}")
        self._assert_dest(b"/base/compilation", item)

    def test_first_different_sep(self, item):
        self._setf("%first{Alice / Bob / Eve,2,0, / , & }")
        self._assert_dest(b"/base/Alice & Bob", item)


class TestDisambiguation(TestHelper, PathFormattingMixin):
    @pytest.fixture(autouse=True)
    def items(self, setup):
        self.lib.directory = b"/base"
        self.lib.path_formats = [("default", "path")]

        i1 = item()
        i1.year = 2001
        self.lib.add_album([i1])
        i2 = item()
        i2.year = 2002
        self.lib.add_album([i2])
        self.lib._connection().commit()

        self._setf("foo%aunique{albumartist album,year}/$title")
        return i1, i2

    def test_unique_expands_to_disambiguating_year(self, items):
        i1, _i2 = items
        self._assert_dest(b"/base/foo [2001]/the title", i1)

    def test_unique_with_default_arguments_uses_albumtype(self, items):
        i1, _i2 = items
        album2 = self.lib.get_album(i1)
        album2.albumtype = "bar"
        album2.store()
        self._setf("foo%aunique{}/$title")
        self._assert_dest(b"/base/foo [bar]/the title", i1)

    def test_unique_expands_to_nothing_for_distinct_albums(self, items):
        i1, i2 = items
        album2 = self.lib.get_album(i2)
        album2.album = "different album"
        album2.store()

        self._assert_dest(b"/base/foo/the title", i1)

    def test_use_fallback_numbers_when_identical(self, items):
        i1, i2 = items
        album2 = self.lib.get_album(i2)
        album2.year = 2001
        album2.store()

        self._assert_dest(b"/base/foo [1]/the title", i1)
        self._assert_dest(b"/base/foo [2]/the title", i2)

    def test_unique_falls_back_to_second_distinguishing_field(self, items):
        i1, _i2 = items
        self._setf("foo%aunique{albumartist album,month year}/$title")
        self._assert_dest(b"/base/foo [2001]/the title", i1)

    def test_unique_sanitized(self, items):
        i1, i2 = items
        album2 = self.lib.get_album(i2)
        album2.year = 2001
        album1 = self.lib.get_album(i1)
        album1.albumtype = "foo/bar"
        album2.store()
        album1.store()
        self._setf("foo%aunique{albumartist album,albumtype}/$title")
        self._assert_dest(b"/base/foo [foo_bar]/the title", i1)

    def test_drop_empty_disambig_string(self, items):
        i1, i2 = items
        album1 = self.lib.get_album(i1)
        album1.albumdisambig = None
        album2 = self.lib.get_album(i2)
        album2.albumdisambig = "foo"
        album1.store()
        album2.store()
        self._setf("foo%aunique{albumartist album,albumdisambig}/$title")
        self._assert_dest(b"/base/foo/the title", i1)

    def test_change_brackets(self, items):
        i1, _i2 = items
        self._setf("foo%aunique{albumartist album,year,()}/$title")
        self._assert_dest(b"/base/foo (2001)/the title", i1)

    def test_remove_brackets(self, items):
        i1, _i2 = items
        self._setf("foo%aunique{albumartist album,year,}/$title")
        self._assert_dest(b"/base/foo 2001/the title", i1)

    def test_key_flexible_attribute(self, items):
        i1, i2 = items
        album1 = self.lib.get_album(i1)
        album1.flex = "flex1"
        album2 = self.lib.get_album(i2)
        album2.flex = "flex2"
        album1.store()
        album2.store()
        self._setf("foo%aunique{albumartist album flex,year}/$title")
        self._assert_dest(b"/base/foo/the title", i1)


class TestSingletonDisambiguation(TestHelper, PathFormattingMixin):
    @pytest.fixture(autouse=True)
    def items(self, setup):
        self.lib.directory = b"/base"
        self.lib.path_formats = [("default", "path")]

        i1 = item()
        i1.year = 2001
        self.lib.add(i1)
        i2 = item()
        i2.year = 2002
        self.lib.add(i2)
        self.lib._connection().commit()

        self._setf("foo/$title%sunique{artist title,year}")
        return i1, i2

    def test_sunique_expands_to_disambiguating_year(self, items):
        i1, _i2 = items
        self._assert_dest(b"/base/foo/the title [2001]", i1)

    def test_sunique_with_default_arguments_uses_trackdisambig(self, items):
        i1, i2 = items
        i1.trackdisambig = "live version"
        i1.year = i2.year
        i1.store()
        self._setf("foo/$title%sunique{}")
        self._assert_dest(b"/base/foo/the title [live version]", i1)

    def test_sunique_expands_to_nothing_for_distinct_singletons(self, items):
        i1, i2 = items
        i2.title = "different track"
        i2.store()

        self._assert_dest(b"/base/foo/the title", i1)

    def test_sunique_does_not_match_album(self, items):
        i1, i2 = items
        self.lib.add_album([i2])
        self._assert_dest(b"/base/foo/the title", i1)

    def test_sunique_use_fallback_numbers_when_identical(self, items):
        i1, i2 = items
        i2.year = i1.year
        i2.store()

        self._assert_dest(b"/base/foo/the title [1]", i1)
        self._assert_dest(b"/base/foo/the title [2]", i2)

    def test_sunique_falls_back_to_second_distinguishing_field(self, items):
        i1, _i2 = items
        self._setf("foo/$title%sunique{albumartist album,month year}")
        self._assert_dest(b"/base/foo/the title [2001]", i1)

    def test_sunique_sanitized(self, items):
        i1, i2 = items
        i2.year = i1.year
        i1.trackdisambig = "foo/bar"
        i2.store()
        i1.store()
        self._setf("foo/$title%sunique{artist title,trackdisambig}")
        self._assert_dest(b"/base/foo/the title [foo_bar]", i1)

    def test_drop_empty_disambig_string(self, items):
        i1, i2 = items
        i1.trackdisambig = None
        i2.trackdisambig = "foo"
        i1.store()
        i2.store()
        self._setf("foo/$title%sunique{albumartist album,trackdisambig}")
        self._assert_dest(b"/base/foo/the title", i1)

    def test_change_brackets(self, items):
        i1, _i2 = items
        self._setf("foo/$title%sunique{artist title,year,()}")
        self._assert_dest(b"/base/foo/the title (2001)", i1)

    def test_remove_brackets(self, items):
        i1, _i2 = items
        self._setf("foo/$title%sunique{artist title,year,}")
        self._assert_dest(b"/base/foo/the title 2001", i1)

    def test_key_flexible_attribute(self, items):
        i1, i2 = items
        i1.flex = "flex1"
        i2.flex = "flex2"
        i1.store()
        i2.store()
        self._setf("foo/$title%sunique{artist title flex,year}")
        self._assert_dest(b"/base/foo/the title", i1)


class TestPluginDestination(TestHelper):
    @pytest.fixture(autouse=True)
    def item(self, setup):
        # Mock beets.plugins.item_field_getters.
        self._tv_map = {}

        def field_getters():
            getters = {}
            for key, value in self._tv_map.items():
                getters[key] = lambda _: value
            return getters

        self.old_field_getters = plugins.item_field_getters
        plugins.item_field_getters = field_getters

        self.lib.directory = b"/base"
        self.lib.path_formats = [("default", "$artist $foo")]

        yield _common.item(self.lib)

        plugins.item_field_getters = self.old_field_getters

    def _assert_dest(self, dest, item):
        with _common.platform_posix():
            the_dest = item.destination()
        assert the_dest == b"/base/" + dest

    def test_undefined_value_not_substituted(self, item):
        self._assert_dest(b"the artist $foo", item)

    def test_plugin_value_not_substituted(self, item):
        self._tv_map = {"foo": "bar"}
        self._assert_dest(b"the artist bar", item)

    def test_plugin_value_overrides_attribute(self, item):
        self._tv_map = {"artist": "bar"}
        self._assert_dest(b"bar $foo", item)

    def test_plugin_value_sanitized(self, item):
        self._tv_map = {"foo": "bar/baz"}
        self._assert_dest(b"the artist bar_baz", item)


class TestAlbumInfo(PytestItemHelper):
    @pytest.fixture
    def item_in_album(self, item):
        self.lib.add_album((item,))
        return item

    def test_albuminfo_reflects_metadata(self, item_in_album):
        ai = self.lib.get_album(item_in_album)
        assert ai.mb_albumartistid == item_in_album.mb_albumartistid
        assert ai.albumartist == item_in_album.albumartist
        assert ai.album == item_in_album.album
        assert ai.year == item_in_album.year

    def test_albuminfo_stores_art(self, item_in_album):
        ai = self.lib.get_album(item_in_album)
        ai.artpath = os.fsdecode(np("/my/great/art"))
        ai.store()
        new_ai = self.lib.get_album(item_in_album)
        assert new_ai.artpath == np("/my/great/art")

    def test_albuminfo_for_two_items_doesnt_duplicate_row(self, item_in_album):
        i2 = _common.item(self.lib)
        self.lib.get_album(item_in_album)
        self.lib.get_album(i2)

        c = self.lib._connection().cursor()
        c.execute("select * from albums where album=?", (item_in_album.album,))
        # Cursor should only return one row.
        assert c.fetchone() is not None
        assert c.fetchone() is None

    def test_individual_tracks_have_no_albuminfo(self):
        i2 = _common.item()
        i2.album = "aTotallyDifferentAlbum"
        self.lib.add(i2)
        ai = self.lib.get_album(i2)
        assert ai is None

    def test_get_album_by_id(self, item_in_album):
        ai = self.lib.get_album(item_in_album)
        ai = self.lib.get_album(item_in_album.id)
        assert ai is not None

    def test_album_items_consistent(self, item_in_album):
        ai = self.lib.get_album(item_in_album)
        assert item_in_album.id in {i.id for i in ai.items()}

    def test_albuminfo_changes_affect_items(self, item_in_album):
        ai = self.lib.get_album(item_in_album)
        ai.album = "myNewAlbum"
        ai.store()
        assert self.get_first_item().album == "myNewAlbum"

    def test_albuminfo_change_albumartist_changes_items(self, item_in_album):
        ai = self.lib.get_album(item_in_album)
        ai.albumartist = "myNewArtist"
        ai.store()
        item = self.get_first_item()
        assert item.albumartist == "myNewArtist"
        assert item.artist != "myNewArtist"

    def test_albuminfo_change_artist_does_change_items(self, item_in_album):
        ai = self.lib.get_album(item_in_album)
        ai.artist = "myNewArtist"
        ai.store(inherit=True)
        assert self.get_first_item().artist == "myNewArtist"

    def test_albuminfo_change_artist_does_not_change_items(self, item_in_album):
        ai = self.lib.get_album(item_in_album)
        ai.artist = "myNewArtist"
        ai.store(inherit=False)
        assert self.get_first_item().artist != "myNewArtist"

    def test_albuminfo_remove_removes_items(self, item_in_album):
        item_id = item_in_album.id
        self.lib.get_album(item_in_album).remove()
        c = self.lib._connection().execute(
            "SELECT id FROM items WHERE id=?", (item_id,)
        )
        assert c.fetchone() is None

    def test_removing_last_item_removes_album(self, item_in_album):
        assert len(self.lib.albums()) == 1
        item_in_album.remove()
        assert len(self.lib.albums()) == 0

    def test_noop_albuminfo_changes_affect_items(self, item_in_album):
        item = self.get_first_item()
        item.album = "foobar"
        item.store()
        ai = self.lib.get_album(item_in_album)
        ai.album = ai.album
        ai.store()
        item = self.get_first_item()
        assert item.album == ai.album


class TestArtDestination(TestHelper):
    @pytest.fixture(autouse=True)
    def item_and_album(self, setup):
        config["art_filename"] = "artimage"
        config["replace"] = {"X": "Y"}
        self.lib.replacements = [(re.compile("X"), "Y")]
        self.lib.path_formats = [("default", "$artist/$album/$track $title")]
        item = _common.item(self.lib)
        item.path = item.destination()
        ai = self.lib.add_album((item,))
        return item, ai

    def test_art_filename_respects_setting(self, item_and_album):
        _i, ai = item_and_album
        art = ai.art_destination("something.jpg")
        new_art = bytestring_path(f"{os.path.sep}artimage.jpg")
        assert new_art in art

    def test_art_path_in_item_dir(self, item_and_album):
        i, ai = item_and_album
        art = ai.art_destination("something.jpg")
        track = i.destination()
        assert os.path.dirname(art) == os.path.dirname(track)

    def test_art_path_sanitized(self, item_and_album):
        _i, ai = item_and_album
        config["art_filename"] = "artXimage"
        art = ai.art_destination("something.jpg")
        assert b"artYimage" in art


class TestPathString(PytestItemHelper):
    def test_item_path_is_bytestring(self, item_in_db):
        assert isinstance(item_in_db.path, bytes)

    def test_fetched_item_path_is_bytestring(self, item_in_db):
        assert isinstance(self.get_first_item().path, bytes)

    def test_unicode_path_becomes_bytestring(self, item_in_db):
        item_in_db.path = "unicodepath"
        assert isinstance(item_in_db.path, bytes)

    def test_unicode_in_database_becomes_bytestring(self, item_in_db):
        self.lib._connection().execute(
            """
        update items set path=? where id=?
        """,
            (item_in_db.id, "somepath"),
        )
        assert isinstance(self.get_first_item().path, bytes)

    def test_special_chars_preserved_in_database(self, item_in_db):
        path = Path("b\xe1r")
        item_in_db.path = path
        item_in_db.store()
        assert self.get_first_item().filepath == self.lib_path / path

    def test_special_char_path_added_to_database(self, item, item_in_db):
        item_in_db.remove()
        path = Path("b\xe1r")
        item = _common.item()
        item.path = path
        self.lib.add(item)
        assert self.get_first_item().filepath == self.lib_path / path

    def test_destination_returns_bytestring(self, item_in_db):
        item_in_db.artist = "b\xe1r"
        dest = item_in_db.destination()
        assert isinstance(dest, bytes)

    def test_art_destination_returns_bytestring(self, item_in_db):
        item_in_db.artist = "b\xe1r"
        alb = self.lib.add_album([item_in_db])
        dest = alb.art_destination("image.jpg")
        assert isinstance(dest, bytes)

    def test_artpath_stores_special_chars(self, item_in_db):
        path = Path("b\xe1r")
        alb = self.lib.add_album([item_in_db])
        alb.artpath = path
        alb.store()
        stored_path = (
            self.lib._connection()
            .execute("select artpath from albums where id=?", (alb.id,))
            .fetchone()[0]
        )
        alb = self.lib.get_album(item_in_db)
        assert stored_path == os.fsencode(path)
        assert alb.art_filepath == self.lib_path / path

    def test_sanitize_path_with_special_chars(self):
        path = "b\xe1r?"
        new_path = util.sanitize_path(path)
        assert new_path.startswith("b\xe1r")

    def test_sanitize_path_returns_unicode(self):
        path = "b\xe1r?"
        new_path = util.sanitize_path(path)
        assert isinstance(new_path, str)

    def test_unicode_artpath_becomes_bytestring(self, item_in_db):
        alb = self.lib.add_album([item_in_db])
        alb.artpath = "somep\xe1th"
        assert isinstance(alb.artpath, bytes)

    def test_unicode_artpath_in_database_decoded(self, item_in_db):
        alb = self.lib.add_album([item_in_db])
        self.lib._connection().execute(
            "update albums set artpath=? where id=?", ("somep\xe1th", alb.id)
        )
        alb = self.lib.get_album(alb.id)
        assert isinstance(alb.artpath, bytes)

    def test_relative_path_is_stored(self, item_in_db):
        relative_path = Path("abc") / "foo.mp3"
        absolute_path = self.lib_path / relative_path
        item_in_db.path = absolute_path
        item_in_db.store()
        stored_path = (
            self.lib._connection()
            .execute("select path from items where id=?", (item_in_db.id,))
            .fetchone()[0]
        )
        album = self.lib.add_album([item_in_db])

        assert item_in_db.filepath == absolute_path
        assert stored_path == util.path_as_posix(os.fsencode(relative_path))
        assert album.filepath == absolute_path.parent


class TestMtime(TestHelper):
    @pytest.fixture(autouse=True)
    def item(self, setup):
        self.ipath = self.temp_path / "testfile.mp3"
        shutil.copy(syspath(_common.RSRC / "full.mp3"), syspath(self.ipath))
        item = beets.library.Item.from_path(self.ipath)
        self.lib.add(item)
        yield item
        if self.ipath.exists():
            self.ipath.unlink()

    def _mtime(self):
        return int(self.ipath.stat().st_mtime)

    def test_mtime_initially_up_to_date(self, item):
        assert item.mtime >= self._mtime()

    def test_mtime_reset_on_db_modify(self, item):
        item.title = "something else"
        assert item.mtime < self._mtime()

    def test_mtime_up_to_date_after_write(self, item):
        item.title = "something else"
        item.write()
        assert item.mtime >= self._mtime()

    def test_mtime_up_to_date_after_read(self, item):
        item.title = "something else"
        item.read()
        assert item.mtime >= self._mtime()


class TestImportTime(TestHelper):
    def added(self):
        self.track = item()
        self.album = self.lib.add_album((self.track,))
        assert self.album.added > 0
        assert self.track.added > 0

    def test_atime_for_singleton(self):
        self.singleton = item(self.lib)
        assert self.singleton.added > 0


class TestTemplate(PytestItemHelper):
    def test_year_formatted_in_template(self, item_in_db):
        item_in_db.year = 123
        item_in_db.store()
        assert item_in_db.evaluate_template("$year") == "0123"

    def test_album_flexattr_appears_in_item_template(self, item_in_db):
        self.album = self.lib.add_album([item_in_db])
        self.album.foo = "baz"
        self.album.store()
        assert item_in_db.evaluate_template("$foo") == "baz"

    def test_album_and_item_format(self, item_in_db):
        config["format_album"] = "foö $foo"
        album = beets.library.Album()
        album.foo = "bar"
        album.tagada = "togodo"
        assert f"{album}" == "foö bar"
        assert f"{album:$tagada}" == "togodo"
        assert str(album) == "foö bar"
        assert bytes(album) == b"fo\xc3\xb6 bar"

        config["format_item"] = "bar $foo"
        item = beets.library.Item()
        item.foo = "bar"
        item.tagada = "togodo"
        assert f"{item}" == "bar bar"
        assert f"{item:$tagada}" == "togodo"


class TestUnicodePath(PytestItemHelper):
    def test_unicode_path(self, item_in_db):
        item_in_db.path = _common.RSRC / "unicode\u2019d.mp3"
        # If there are any problems with unicode paths, we will raise
        # here and fail.
        item_in_db.read()
        item_in_db.write()


class TestWrite(TestHelper):
    def test_write_nonexistant(self):
        item = self.create_item()
        item.path = b"/path/does/not/exist"
        with pytest.raises(beets.library.ReadError):
            item.write()

    def test_no_write_permission(self):
        item = self.add_item_fixture()
        path = syspath(item.path)
        os.chmod(path, stat.S_IRUSR)

        try:
            with pytest.raises(beets.library.WriteError) as exc_info:
                item.write()
            assert "super:" not in str(exc_info.value)

        finally:
            # Restore write permissions so the file can be cleaned up.
            os.chmod(path, stat.S_IRUSR | stat.S_IWUSR)

    def test_write_with_custom_path(self):
        item = self.add_item_fixture()
        custom_path = self.temp_path / "custom.mp3"
        shutil.copy(syspath(item.path), syspath(custom_path))

        item["artist"] = "new artist"
        assert MediaFile(syspath(custom_path)).artist != "new artist"
        assert MediaFile(syspath(item.path)).artist != "new artist"

        item.write(custom_path)
        assert MediaFile(syspath(custom_path)).artist == "new artist"
        assert MediaFile(syspath(item.path)).artist != "new artist"

    def test_write_custom_tags(self):
        item = self.add_item_fixture(artist="old artist")
        item.write(tags={"artist": "new artist"})
        assert item.artist != "new artist"
        assert MediaFile(syspath(item.path)).artist == "new artist"

    def test_write_multi_tags(self):
        item = self.add_item_fixture(artist="old artist")
        item.write(tags={"artists": ["old artist", "another artist"]})

        assert MediaFile(syspath(item.path)).artists == [
            "old artist",
            "another artist",
        ]

    def test_write_multi_tags_id3v23(self):
        item = self.add_item_fixture(artist="old artist")
        item.write(
            tags={"artists": ["old artist", "another artist"]}, id3v23=True
        )

        assert MediaFile(syspath(item.path)).artists == [
            "old artist/another artist"
        ]

    def test_write_date_field(self):
        # Since `date` is not a MediaField, this should do nothing.
        item = self.add_item_fixture()
        clean_year = item.year
        item.date = "foo"
        item.write()
        assert MediaFile(syspath(item.path)).year == clean_year


class TestItemRead(PytestItemHelper):
    def test_unreadable_raise_read_error(self, item_in_db):
        unreadable = _common.RSRC / "image-2x3.png"
        with pytest.raises(beets.library.ReadError) as exc_info:
            item_in_db.read(unreadable)
        assert isinstance(exc_info.value.reason, UnreadableFileError)

    def test_nonexistent_raise_read_error(self, item_in_db):
        with pytest.raises(beets.library.ReadError):
            item_in_db.read("/thisfiledoesnotexist")

    def test_read_error_str_includes_reason(self, item_in_db):
        unreadable = _common.RSRC / "image-2x3.png"
        with pytest.raises(beets.library.ReadError) as exc_info:
            item_in_db.read(unreadable)
        message = str(exc_info.value)
        assert "super:" not in message
        assert str(exc_info.value.reason) in message


class TestItemReadGenre(TestHelper):
    def test_read_semicolon_delimited_genres(self):
        """Semicolon-delimited genre tags are split into individual genres on read."""
        path = self.create_mediafile_fixture()
        mf = MediaFile(syspath(path))
        mf.genres = ["Jazz; Funk; Soul"]
        mf.save()
        item = beets.library.Item.from_path(path)
        assert item.genres == ["Jazz", "Funk", "Soul"]


class TestFilesize(TestHelper):
    def test_filesize(self):
        item = self.add_item_fixture()
        assert item.filesize != 0

    def test_nonexistent_file(self):
        item = beets.library.Item()
        assert item.filesize == 0


class TestItemPruneDirsClutter(TestHelper):
    """Regression tests: prune_dirs respects config["clutter"] during move/remove."""

    def _drop_clutter(self, directory, filename=b"unwanted.log"):
        """Create a clutter file in *directory* (bytes path)."""
        path = os.path.join(directory, filename)
        with open(syspath(path), "w"):
            pass
        return path

    def test_move_prunes_dir_with_config_clutter(self):
        """After moving an item, old dir is removed even when only clutter remains."""
        config["clutter"] = ["*.log"]
        item = self.add_item_fixture()
        old_dir = os.path.dirname(item.path)
        self._drop_clutter(old_dir)

        # Change artist so the destination path differs, forcing a real move.
        item.artist = "new artist"
        item.store()
        item.move()

        assert not os.path.exists(syspath(old_dir))

    def test_remove_prunes_dir_with_config_clutter(self):
        """After deleting an item, its dir is removed even when only clutter remains."""
        config["clutter"] = ["*.log"]
        item = self.add_item_fixture()
        old_dir = os.path.dirname(item.path)
        self._drop_clutter(old_dir)

        item.remove(delete=True)

        assert not os.path.exists(syspath(old_dir))


class TestParseQuery:
    def test_parse_invalid_query_string(self):
        with pytest.raises(beets.dbcore.query.ParsingError):
            beets.library.parse_query_string('foo"', None)

    def test_parse_bytes(self):
        with pytest.raises(AssertionError):
            beets.library.parse_query_string(b"query", None)



import os
import subprocess
import sys

import pytest

from beets.test import _common
from beets.test.helper import RUNNING_IN_CI, IOMixin, has_program
from beets.ui.commands.completion import BASH_COMPLETION_PATHS
from beets.util import syspath

from ..test_ui import TestPluginTestCase


@pytest.mark.xfail(
    RUNNING_IN_CI and sys.platform == "linux",
    reason="Completion is for some reason unhappy on Ubuntu 24.04 in CI",
)
class CompletionTest(IOMixin, TestPluginTestCase):
    def test_completion(self):
        # Do not load any other bash completion scripts on the system.
        env = dict(os.environ)
        env["BASH_COMPLETION_DIR"] = os.devnull
        env["BASH_COMPLETION_COMPAT_DIR"] = os.devnull

        # Open a `bash` process to run the tests in. We'll pipe in bash
        # commands via stdin.
        cmd = os.environ.get("BEETS_TEST_SHELL", "/bin/bash --norc").split()
        if not has_program(cmd[0]):
            self.skipTest("bash not available")
        tester = subprocess.Popen(
            cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, env=env
        )

        # Load bash_completion library.
        for path in BASH_COMPLETION_PATHS:
            if os.path.exists(syspath(path)):
                bash_completion = path
                break
        else:
            self.skipTest("bash-completion script not found")
        try:
            with open(syspath(bash_completion), "rb") as f:
                tester.stdin.writelines(f)
        except OSError:
            self.skipTest("could not read bash-completion script")

        # Load completion script.
        self.run_command("completion")
        completion_script = self.io.getoutput().encode("utf-8")
        tester.stdin.writelines(completion_script.splitlines(True))

        # Load test suite.
        test_script_name = _common.RSRC / "test_completion.sh"
        with test_script_name.open("rb") as test_script_file:
            tester.stdin.writelines(test_script_file)
        out, _ = tester.communicate()
        assert tester.returncode == 0
        assert out == b"completion tests passed\n", (
            "test/test_completion.sh did not execute properly. "
            f"Output:{out.decode('utf-8')}"
        )



"""Allows custom commands to be run when an event is emitted by beets"""

from __future__ import annotations

import os
import shlex
import string
import subprocess
from typing import Any

from beets.plugins import BeetsPlugin


class BytesToStrFormatter(string.Formatter):
    """A variant of `string.Formatter` that converts `bytes` to `str`."""

    def convert_field(self, value: Any, conversion: str | None) -> Any:
        """Converts the provided value given a conversion type.

        This method decodes the converted value using the formatter's coding.
        """
        converted = super().convert_field(value, conversion)

        if isinstance(converted, bytes):
            return os.fsdecode(converted)

        return converted


class HookPlugin(BeetsPlugin):
    """Allows custom commands to be run when an event is emitted by beets"""

    def __init__(self):
        super().__init__()

        self.config.add({"hooks": []})

        hooks = self.config["hooks"].get(list)

        for hook_index in range(len(hooks)):
            hook = self.config["hooks"][hook_index]

            hook_event = hook["event"].as_str()
            hook_command = hook["command"].as_str()

            self.create_and_register_hook(hook_event, hook_command)

    def create_and_register_hook(self, event, command):
        def hook_function(**kwargs):
            if command is None or len(command) == 0:
                self._log.error('invalid command "{}"', command)
                return

            # For backwards compatibility, use a string formatter that decodes
            # bytes (in particular, paths) to strings.
            formatter = BytesToStrFormatter()
            command_pieces = [
                formatter.format(piece, event=event, **kwargs)
                for piece in shlex.split(command)
            ]

            self._log.debug(
                'running command "{}" for event {}',
                " ".join(command_pieces),
                event,
            )

            try:
                subprocess.check_call(command_pieces)
            except subprocess.CalledProcessError as exc:
                self._log.error(
                    "hook for {} exited with status {.returncode}", event, exc
                )
            except OSError as exc:
                self._log.error("hook for {} failed: {}", event, exc)

        self.register_listener(event, hook_function)



from __future__ import annotations

from typing import TYPE_CHECKING

from .functemplate import template

if TYPE_CHECKING:
    import confuse

    from .functemplate import Template

    PathFormat = tuple[str, Template]


# Special path format key.
PF_KEY_DEFAULT = "default"
PF_KEY_QUERIES = {"comp": "comp:true", "singleton": "singleton:true"}


def get_path_formats(subview: confuse.Subview) -> list[PathFormat]:
    """Build ordered query/template pairs from layered path-format settings.

    The mapping is read through Confuse's ``items()`` view so keys from lower-
    priority sources remain visible when higher-priority config only overrides
    part of ``paths``. This keeps inherited defaults such as ``default``,
    ``comp``, and ``singleton`` available unless they are explicitly replaced.
    """
    return [
        (PF_KEY_QUERIES.get(q, q), template(v.as_str()))
        for q, v in subview.items()
    ]



from __future__ import annotations

from enum import Enum

from typing_extensions import Self


class Action(Enum):
    """Enumeration of possible actions for an import task."""

    SKIP = "SKIP"
    ASIS = "ASIS"
    TRACKS = "TRACKS"
    APPLY = "APPLY"
    ALBUMS = "ALBUMS"
    RETAG = "RETAG"
    # The RETAG action represents "don't apply any match, but do record
    # new metadata". It's not reachable via the standard command prompt but
    # can be used by plugins.


class DuplicateAction(str, Enum):
    text: str

    def __new__(cls, code: str, text: str) -> Self:
        obj = str.__new__(cls, code)
        obj._value_ = code
        obj.text = text
        return obj

    @classmethod
    def options(cls) -> list[str]:
        return [d.text for d in cls]

    @classmethod
    def strict_options(cls) -> list[str]:
        return [d.text for d in set(cls) - {DuplicateAction.ASK}]

    @classmethod
    def choices(cls) -> dict[str, str]:
        return {d.name.lower(): d.value for d in cls}

    SKIP = "s", "Skip new"
    MERGE = "m", "Merge all"
    REMOVE = "r", "Remove old"
    KEEP = "k", "Keep all"
    ASK = "a", "Ask"



"""Tests for the 'limit' plugin."""

from beets.test.helper import IOMixin, PluginTestCase


class LimitPluginTest(IOMixin, PluginTestCase):
    """Unit tests for LimitPlugin

    Note: query prefix tests do not work correctly with `run_with_output`.
    """

    plugin = "limit"

    def setUp(self):
        super().setUp()

        # we'll create an even number of tracks in the library
        self.num_test_items = 10
        assert self.num_test_items % 2 == 0
        for item_no, item in enumerate(
            self.add_item_fixtures(count=self.num_test_items)
        ):
            item.track = item_no + 1
            item.store()

        # our limit tests will use half of this number
        self.num_limit = self.num_test_items // 2
        self.num_limit_prefix = "".join(["'", "<", str(self.num_limit), "'"])

        # a subset of tests has only `num_limit` results, identified by a
        # range filter on the track number
        self.track_head_range = f"track:..{self.num_limit}"
        self.track_tail_range = f"track:{self.num_limit + 1}{'..'}"

    def test_no_limit(self):
        """Returns all when there is no limit or filter."""
        result = self.run_with_output("lslimit")
        assert result.count("\n") == self.num_test_items

    def test_lslimit_head(self):
        """Returns the expected number with `lslimit --head`."""
        result = self.run_with_output("lslimit", "--head", str(self.num_limit))
        assert result.count("\n") == self.num_limit

    def test_lslimit_tail(self):
        """Returns the expected number with `lslimit --tail`."""
        result = self.run_with_output("lslimit", "--tail", str(self.num_limit))
        assert result.count("\n") == self.num_limit

    def test_lslimit_head_invariant(self):
        """Returns the expected number with `lslimit --head` and a filter."""
        result = self.run_with_output(
            "lslimit", "--head", str(self.num_limit), self.track_tail_range
        )
        assert result.count("\n") == self.num_limit

    def test_lslimit_tail_invariant(self):
        """Returns the expected number with `lslimit --tail` and a filter."""
        result = self.run_with_output(
            "lslimit", "--tail", str(self.num_limit), self.track_head_range
        )
        assert result.count("\n") == self.num_limit

    def test_prefix(self):
        """Returns the expected number with the query prefix."""
        result = self.lib.items(self.num_limit_prefix)
        assert len(result) == self.num_limit

    def test_prefix_when_correctly_ordered(self):
        """Returns the expected number with the query prefix and filter when
        the prefix portion (correctly) appears last."""
        correct_order = f"{self.track_tail_range} {self.num_limit_prefix}"
        result = self.lib.items(correct_order)
        assert len(result) == self.num_limit

    def test_prefix_when_incorrectly_ordred(self):
        """Returns no results with the query prefix and filter when the prefix
        portion (incorrectly) appears first."""
        incorrect_order = f"{self.num_limit_prefix} {self.track_tail_range}"
        result = self.lib.items(incorrect_order)
        assert len(result) == 0



import shutil

import pytest

from beets import library
from beets.exceptions import UserError
from beets.test import _common
from beets.test.helper import BeetsTestCase
from beets.ui.commands.utils import do_query
from beets.util import syspath


class QueryTest(BeetsTestCase):
    def add_item(self):
        itempath = self.lib_path / "srcfile"
        shutil.copy(_common.RSRC / "full.mp3", syspath(itempath))
        item = library.Item.from_path(itempath)
        self.lib.add(item)
        return item

    def add_album(self, items):
        return self.lib.add_album(items)

    def check_do_query(
        self, num_items, num_albums, q=(), album=False, also_items=True
    ):
        items, albums = do_query(self.lib, q, album, also_items)
        assert len(items) == num_items
        assert len(albums) == num_albums

    def test_query_empty(self):
        with pytest.raises(UserError):
            do_query(self.lib, (), False)

    def test_query_empty_album(self):
        with pytest.raises(UserError):
            do_query(self.lib, (), True)

    def test_query_item(self):
        self.add_item()
        self.check_do_query(1, 0, album=False)
        self.add_item()
        self.check_do_query(2, 0, album=False)

    def test_query_album(self):
        item = self.add_item()
        self.add_album([item])
        self.check_do_query(1, 1, album=True)
        self.check_do_query(0, 1, album=True, also_items=False)

        item = self.add_item()
        item2 = self.add_item()
        self.add_album([item, item2])
        self.check_do_query(3, 2, album=True)
        self.check_do_query(0, 2, album=True, also_items=False)



from __future__ import annotations

import os
import os.path
import shutil
import tempfile
import unittest
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import pytest
from mediafile import MediaFile

from beets import config, logging
from beets.exceptions import UserError
from beets.test import _common
from beets.test.fixtures import DummyIMBackend
from beets.test.helper import (
    NEEDS_FFPROBE,
    FetchImageHelper,
    ImportHelper,
    IOMixin,
    PluginMixin,
)
from beets.util import bytestring_path, displayable_path, syspath
from beets.util.artresizer import ArtResizer
from beetsplug._utils import art

if TYPE_CHECKING:
    from beets.test.helper import ImageRequestMocker


def require_artresizer_compare(test):
    def wrapper(*args, **kwargs):
        if not ArtResizer.shared.can_compare:
            raise unittest.SkipTest("compare not available")

        # PHASH computation in ImageMagick changed at some point in an
        # undocumented way. Check at a low level that comparisons of our
        # fixtures give the expected results. Only then, plugin logic tests
        # below are meaningful.
        # cf. https://github.com/ImageMagick/ImageMagick/discussions/5191
        # It would be better to investigate what exactly change in IM and
        # handle that in ArtResizer.IMBackend.{can_compare,compare}.
        # Skipping the tests as below is a quick fix to CI, but users may
        # still see unexpected behaviour.
        abbey_artpath = _common.RSRC / "abbey.jpg"
        abbey_similarpath = _common.RSRC / "abbey-similar.jpg"
        abbey_differentpath = _common.RSRC / "abbey-different.jpg"
        compare_threshold = 20

        similar_compares_ok = ArtResizer.shared.compare(
            abbey_artpath, abbey_similarpath, compare_threshold
        )
        different_compares_ok = ArtResizer.shared.compare(
            abbey_artpath, abbey_differentpath, compare_threshold
        )
        if not similar_compares_ok or different_compares_ok:
            raise unittest.SkipTest("IM version with broken compare")

        return test(*args, **kwargs)

    wrapper.__name__ = test.__name__
    return wrapper


class TestEmbedartCli(PluginMixin, IOMixin, ImportHelper, FetchImageHelper):
    plugin = "embedart"
    small_artpath = _common.RSRC / "image-2x3.jpg"
    abbey_artpath = _common.RSRC / "abbey.jpg"
    abbey_similarpath = _common.RSRC / "abbey-similar.jpg"
    abbey_differentpath = _common.RSRC / "abbey-different.jpg"

    def _setup_data(self, artpath=None):
        if not artpath:
            artpath = self.small_artpath
        with open(syspath(artpath), "rb") as f:
            self.image_data = f.read()

    def test_embed_art_from_file_with_yes_input(self):
        self._setup_data()
        album = self.add_album_fixture()
        item = album.items()[0]
        self.io.addinput("y")
        self.run_command("embedart", "-f", self.small_artpath)
        mediafile = MediaFile(syspath(item.path))
        assert mediafile.images[0].data == self.image_data

    def test_embed_art_from_file_with_no_input(self):
        self._setup_data()
        album = self.add_album_fixture()
        item = album.items()[0]
        self.io.addinput("n")
        self.run_command("embedart", "-f", self.small_artpath)
        mediafile = MediaFile(syspath(item.path))
        # make sure that images array is empty (nothing embedded)
        assert not mediafile.images

    def test_embed_art_from_file(self):
        self._setup_data()
        album = self.add_album_fixture()
        item = album.items()[0]
        self.run_command("embedart", "-y", "-f", self.small_artpath)
        mediafile = MediaFile(syspath(item.path))
        assert mediafile.images[0].data == self.image_data

    def test_embed_art_from_album(self):
        self._setup_data()
        album = self.add_album_fixture()
        item = album.items()[0]
        album.artpath = self.small_artpath
        album.store()
        self.run_command("embedart", "-y")
        mediafile = MediaFile(syspath(item.path))
        assert mediafile.images[0].data == self.image_data

    def test_embed_art_remove_art_file(self):
        self._setup_data()
        album = self.add_album_fixture()

        logging.getLogger("beets.embedart").setLevel(logging.DEBUG)

        handle, tmp_path = tempfile.mkstemp()
        tmp_path = bytestring_path(tmp_path)
        os.write(handle, self.image_data)
        os.close(handle)

        album.artpath = tmp_path
        album.store()

        config["embedart"]["remove_art_file"] = True
        self.run_command("embedart", "-y")

        if os.path.isfile(syspath(tmp_path)):
            os.remove(syspath(tmp_path))
            pytest.fail(
                f"Artwork file {displayable_path(tmp_path)} was not deleted"
            )

    def test_art_file_missing(self):
        self.add_album_fixture()
        logging.getLogger("beets.embedart").setLevel(logging.DEBUG)
        with pytest.raises(UserError):
            self.run_command("embedart", "-y", "-f", "/doesnotexist")

    def test_embed_non_image_file(self):
        album = self.add_album_fixture()
        logging.getLogger("beets.embedart").setLevel(logging.DEBUG)

        handle, tmp_path = tempfile.mkstemp()
        tmp_path = bytestring_path(tmp_path)
        os.write(handle, b"I am not an image.")
        os.close(handle)

        try:
            self.run_command("embedart", "-y", "-f", tmp_path)
        finally:
            os.remove(syspath(tmp_path))

        mediafile = MediaFile(syspath(album.items()[0].path))
        assert not mediafile.images  # No image added.

    @require_artresizer_compare
    def test_reject_different_art(self):
        self._setup_data(self.abbey_artpath)
        album = self.add_album_fixture()
        item = album.items()[0]
        self.run_command("embedart", "-y", "-f", self.abbey_artpath)
        config["embedart"]["compare_threshold"] = 20
        self.run_command("embedart", "-y", "-f", self.abbey_differentpath)
        mediafile = MediaFile(syspath(item.path))

        assert mediafile.images[0].data == self.image_data, (
            f"Image written is not {self.abbey_artpath}"
        )

    @require_artresizer_compare
    def test_accept_similar_art(self):
        self._setup_data(self.abbey_similarpath)
        album = self.add_album_fixture()
        item = album.items()[0]
        self.run_command("embedart", "-y", "-f", self.abbey_artpath)
        config["embedart"]["compare_threshold"] = 20
        self.run_command("embedart", "-y", "-f", self.abbey_similarpath)
        mediafile = MediaFile(syspath(item.path))

        assert mediafile.images[0].data == self.image_data, (
            f"Image written is not {self.abbey_similarpath}"
        )

    def test_non_ascii_album_path(self):
        resource_path = _common.RSRC / "image.mp3"
        album = self.add_album_fixture()
        trackpath = album.items()[0].path
        shutil.copy(syspath(resource_path), syspath(trackpath))

        self.run_command("extractart", "-n", "extracted")

        assert (album.filepath / "extracted.png").exists()

    def test_extracted_extension(self):
        resource_path = _common.RSRC / "image-jpeg.mp3"
        album = self.add_album_fixture()
        trackpath = album.items()[0].path
        shutil.copy(syspath(resource_path), syspath(trackpath))

        self.run_command("extractart", "-n", "extracted")

        assert (album.filepath / "extracted.jpg").exists()

    def test_clear_art_with_yes_input(self):
        self._setup_data()
        album = self.add_album_fixture()
        item = album.items()[0]
        self.io.addinput("y")
        self.run_command("embedart", "-f", self.small_artpath)
        embedded_time = os.path.getmtime(syspath(item.path))

        self.io.addinput("y")
        self.run_command("clearart")
        mediafile = MediaFile(syspath(item.path))
        assert not mediafile.images
        clear_time = os.path.getmtime(syspath(item.path))
        assert clear_time > embedded_time

        # A run on a file without an image should not be modified
        self.io.addinput("y")
        self.run_command("clearart")
        no_clear_time = os.path.getmtime(syspath(item.path))
        assert no_clear_time == clear_time

    def test_clear_art_with_no_input(self):
        self._setup_data()
        album = self.add_album_fixture()
        item = album.items()[0]
        self.io.addinput("y")
        self.run_command("embedart", "-f", self.small_artpath)
        self.io.addinput("n")
        self.run_command("clearart")
        mediafile = MediaFile(syspath(item.path))
        assert mediafile.images[0].data == self.image_data

    def test_embed_art_from_url_with_yes_input(
        self, image_request_mock: ImageRequestMocker
    ):
        self._setup_data()
        album = self.add_album_fixture()
        item = album.items()[0]
        image_request_mock.get(
            "http://example.com/test.jpg", content_type="image/jpeg"
        )
        self.io.addinput("y")
        self.run_command("embedart", "-u", "http://example.com/test.jpg")
        mediafile = MediaFile(syspath(item.path))
        assert mediafile.images[0].data == image_request_mock.IMAGE_HEADERS[
            "image/jpeg"
        ].ljust(32, b"\x00")

    def test_embed_art_from_url_png(
        self, image_request_mock: ImageRequestMocker
    ):
        self._setup_data()
        album = self.add_album_fixture()
        item = album.items()[0]
        image_request_mock.get(
            "http://example.com/test.png", content_type="image/png"
        )
        self.run_command("embedart", "-y", "-u", "http://example.com/test.png")
        mediafile = MediaFile(syspath(item.path))
        assert mediafile.images[0].data == image_request_mock.IMAGE_HEADERS[
            "image/png"
        ].ljust(32, b"\x00")

    def test_embed_art_from_url_not_image(
        self, image_request_mock: ImageRequestMocker
    ):
        self._setup_data()
        album = self.add_album_fixture()
        item = album.items()[0]
        image_request_mock.get(
            "http://example.com/test.html", content_type="text/html"
        )
        self.run_command("embedart", "-y", "-u", "http://example.com/test.html")
        mediafile = MediaFile(syspath(item.path))
        assert not mediafile.images

    @NEEDS_FFPROBE
    def test_clearart_on_import_disabled(self):
        file_path = self.create_mediafile_fixture(
            images=["jpg"], target_dir=self.import_path
        )
        self.import_media.append(file_path)
        with self.configure_plugin({"clearart_on_import": False}):
            importer = self.setup_importer(autotag=False, write=True)
            importer.run()

        item = self.lib.items()[0]
        assert MediaFile(os.path.join(item.path)).images

    @NEEDS_FFPROBE
    def test_clearart_on_import_enabled(self):
        file_path = self.create_mediafile_fixture(
            images=["jpg"], target_dir=self.import_path
        )
        self.import_media.append(file_path)
        # Force re-init the plugin to register the listener
        self.unload_plugins()
        with self.configure_plugin({"clearart_on_import": True}):
            importer = self.setup_importer(autotag=False, write=True)
            importer.run()

        item = self.lib.items()[0]
        assert not MediaFile(os.path.join(item.path)).images


class DummyArtResizer(ArtResizer):
    """An `ArtResizer` which pretends that ImageMagick is available, and has
    a sufficiently recent version to support image comparison.
    """

    def __init__(self):
        self.local_method = DummyIMBackend()


@patch("beets.util.artresizer.subprocess")
@patch("beetsplug._utils.art.extract")
class ArtSimilarityTest(unittest.TestCase):
    def setUp(self):
        self.item = _common.item()
        self.log = logging.getLogger("beets.embedart")
        self.artresizer = DummyArtResizer()

    def _similarity(self, threshold):
        return art.check_art_similarity(
            self.log, self.item, b"path", threshold, artresizer=self.artresizer
        )

    def _popen(self, status=0, stdout="", stderr=""):
        """Create a mock `Popen` object."""
        popen = MagicMock(returncode=status)
        popen.communicate.return_value = stdout, stderr
        return popen

    def _mock_popens(
        self,
        mock_extract,
        mock_subprocess,
        compare_status=0,
        compare_stdout=b"",
        compare_stderr=b"",
        convert_status=0,
    ):
        mock_extract.return_value = b"extracted_path"
        mock_subprocess.Popen.side_effect = [
            # The `convert` call.
            self._popen(convert_status),
            # The `compare` call.
            self._popen(compare_status, compare_stdout, compare_stderr),
        ]

    def test_compare_success_similar(self, mock_extract, mock_subprocess):
        self._mock_popens(mock_extract, mock_subprocess, 0, b"10", b"err")
        assert self._similarity(20)

    def test_compare_success_different(self, mock_extract, mock_subprocess):
        self._mock_popens(mock_extract, mock_subprocess, 0, b"10", b"err")
        assert not self._similarity(5)

    def test_compare_status1_similar(self, mock_extract, mock_subprocess):
        self._mock_popens(mock_extract, mock_subprocess, 1, b"out", b"10")
        assert self._similarity(20)

    def test_compare_status1_different(self, mock_extract, mock_subprocess):
        self._mock_popens(mock_extract, mock_subprocess, 1, b"out", b"10")
        assert not self._similarity(5)

    def test_compare_failed(self, mock_extract, mock_subprocess):
        self._mock_popens(mock_extract, mock_subprocess, 2, b"out", b"10")
        assert self._similarity(20) is None

    def test_compare_parsing_error(self, mock_extract, mock_subprocess):
        self._mock_popens(mock_extract, mock_subprocess, 0, b"foo", b"bar")
        assert self._similarity(20) is None

    def test_compare_parsing_error_and_failure(
        self, mock_extract, mock_subprocess
    ):
        self._mock_popens(mock_extract, mock_subprocess, 1, b"foo", b"bar")
        assert self._similarity(20) is None

    def test_convert_failure(self, mock_extract, mock_subprocess):
        self._mock_popens(mock_extract, mock_subprocess, convert_status=1)
        assert self._similarity(20) is None



import pytest

from beets.util.units import human_bytes, human_seconds


@pytest.mark.parametrize(
    "input_bytes,expected",
    [
        (0, "0.0 B"),
        (30, "30.0 B"),
        (pow(2, 10), "1.0 KiB"),
        (pow(2, 20), "1.0 MiB"),
        (pow(2, 30), "1.0 GiB"),
        (pow(2, 40), "1.0 TiB"),
        (pow(2, 50), "1.0 PiB"),
        (pow(2, 60), "1.0 EiB"),
        (pow(2, 70), "1.0 ZiB"),
        (pow(2, 80), "1.0 YiB"),
        (pow(2, 90), "1.0 HiB"),
        (pow(2, 100), "big"),
    ],
)
def test_human_bytes(input_bytes, expected):
    assert human_bytes(input_bytes) == expected


@pytest.mark.parametrize(
    "input_seconds,expected",
    [
        (0, "0.0 seconds"),
        (30, "30.0 seconds"),
        (60, "1.0 minutes"),
        (90, "1.5 minutes"),
        (125, "2.1 minutes"),
        (3600, "1.0 hours"),
        (86400, "1.0 days"),
        (604800, "1.0 weeks"),
        (31449600, "1.0 years"),
        (314496000, "1.0 decades"),
    ],
)
def test_human_seconds(input_seconds, expected):
    assert human_seconds(input_seconds) == expected



"""The 'config' command: show and edit user configuration."""

import os

from beets import config, ui
from beets.exceptions import UserError
from beets.util import displayable_path, editor_command, interactive_open


def config_func(lib, opts, args):
    # Make sure lazy configuration is loaded
    config.resolve()

    # Print paths.
    if opts.paths:
        filenames = []
        for source in config.sources:
            if not opts.defaults and source.default:
                continue
            if source.filename:
                filenames.append(source.filename)

        # In case the user config file does not exist, prepend it to the
        # list.
        user_path = config.user_config_path()
        if user_path not in filenames:
            filenames.insert(0, user_path)

        for filename in filenames:
            ui.print_(displayable_path(filename))

    # Open in editor.
    elif opts.edit:
        # Note:  This branch *should* be unreachable
        # since the normal flow should be short-circuited
        # by the special case in ui._raw_main
        config_edit(opts)

    # Dump configuration.
    else:
        config_out = config.dump(full=opts.defaults, redact=opts.redact)
        if config_out.strip() != "{}":
            ui.print_(config_out)
        else:
            print("Empty configuration")


def config_edit(cli_options):
    """Open a program to edit the user configuration.
    An empty config file is created if no existing config file exists.
    """
    path = cli_options.config or config.user_config_path()
    editor = editor_command()

    if not editor:
        raise UserError(
            "Please set the VISUAL or EDITOR environment variable to edit"
            " configuration."
        )
    try:
        if not os.path.isfile(path):
            open(path, "w+").close()
        interactive_open([path], editor)
    except FileNotFoundError:
        raise UserError(f"Editor {editor!r} not found.")
    except OSError as exc:
        raise UserError(f"Could not edit configuration: {exc}")


config_cmd = ui.Subcommand("config", help="show or edit the user configuration")
config_cmd.parser.add_option(
    "-p",
    "--paths",
    action="store_true",
    help="show files that configuration was loaded from",
)
config_cmd.parser.add_option(
    "-e",
    "--edit",
    action="store_true",
    help="edit user configuration with $VISUAL (or $EDITOR)",
)
config_cmd.parser.add_option(
    "-d",
    "--defaults",
    action="store_true",
    help="include the default configuration",
)
config_cmd.parser.add_option(
    "-c",
    "--clear",
    action="store_false",
    dest="redact",
    default=True,
    help="do not redact sensitive fields",
)
config_cmd.func = config_func



"""Matches existing metadata with canonical information to identify
releases and tracks.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum
from functools import cached_property
from typing import TYPE_CHECKING, Any, ClassVar, NamedTuple, TypeVar

import lap
import numpy as np

from beets import config, logging, metadata_plugins, plugins
from beets.util import get_most_common_tags

from .distance import VA_ARTISTS, distance, track_distance

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from beets.library import Album, Item

    from .distance import Distance
    from .hooks import AlbumInfo, Info, TrackInfo

    JSONDict = dict[str, Any]
    AnyMatch = TypeVar("AnyMatch", "TrackMatch", "AlbumMatch")
    Candidates = dict[Info.Identifier, AnyMatch]

# Global logger.
log = logging.getLogger("beets")


# Recommendation enumeration.


class Recommendation(IntEnum):
    """Indicates a qualitative suggestion to the user about what should
    be done with a given match.
    """

    none = 0
    low = 1
    medium = 2
    strong = 3


# Primary matching functionality.


def assign_items(
    items: Sequence[Item], tracks: Sequence[TrackInfo]
) -> tuple[list[tuple[Item, TrackInfo]], list[Item], list[TrackInfo]]:
    """Given a list of Items and a list of TrackInfo objects, find the
    best mapping between them. Returns a mapping from Items to TrackInfo
    objects, a set of extra Items, and a set of extra TrackInfo
    objects. These "extra" objects occur when there is an unequal number
    of objects of the two types.
    """
    log.debug("Computing track assignment...")
    # Construct the cost matrix.
    costs = [[float(track_distance(i, t)) for t in tracks] for i in items]
    # Assign items to tracks
    _, _, assigned_item_idxs = lap.lapjv(np.array(costs), extend_cost=True)
    log.debug("...done.")

    # Each item in `assigned_item_idxs` list corresponds to a track in the
    # `tracks` list. Each value is either an index into the assigned item in
    # `items` list, or -1 if that track has no match.
    mapping = {
        items[iidx]: t
        for iidx, t in zip(assigned_item_idxs, tracks)
        if iidx != -1
    }
    extra_items = list(set(items) - mapping.keys())
    extra_items.sort(key=lambda i: (i.disc, i.track, i.title))
    extra_tracks = list(set(tracks) - set(mapping.values()))
    extra_tracks.sort(key=lambda t: (t.index, t.title))
    return list(mapping.items()), extra_items, extra_tracks


# Structures that compose all the information for a candidate match.
@dataclass
class Match:
    """Represent a chosen metadata candidate and its application behavior."""

    disambig_fields_key: ClassVar[str]

    distance: Distance
    info: Info

    def apply_metadata(self) -> None:
        """Apply this match's metadata to its target library objects."""
        raise NotImplementedError

    @cached_property
    def type(self) -> str:
        return self.info.type

    @cached_property
    def config_from_scratch(self) -> bool:
        return bool(config["import"]["from_scratch"])

    def from_scratch(self, override: bool | None) -> bool:
        if override is not None:
            return override

        return self.config_from_scratch

    @property
    def disambig_fields(self) -> Sequence[str]:
        """Return configured disambiguation fields that exist on this match."""
        chosen_fields = config["match"][self.disambig_fields_key].as_str_seq()
        valid_fields = [f for f in chosen_fields if f in self.info]
        if missing_fields := set(chosen_fields) - set(valid_fields):
            log.warning(
                "Disambiguation string keys {} do not exist.", missing_fields
            )

        return valid_fields

    @property
    def base_disambig_data(self) -> JSONDict:
        """Return supplemental values used when formatting disambiguation."""
        return {}

    @property
    def disambig_string(self) -> str:
        """Build a display string from the candidate's disambiguation fields.

        Merges base disambiguation data with instance-specific field values,
        then formats them as a comma-separated string in field definition order.
        """
        data = {
            k: self.info[k] for k in self.disambig_fields
        } | self.base_disambig_data
        return ", ".join(str(data[k]) for k in self.disambig_fields)


@dataclass
class AlbumMatch(Match):
    """Represent an album candidate together with its item-to-track mapping."""

    disambig_fields_key = "album_disambig_fields"

    info: AlbumInfo
    mapping: dict[Item, TrackInfo]
    extra_items: list[Item] = field(default_factory=list)
    extra_tracks: list[TrackInfo] = field(default_factory=list)

    def __post_init__(self) -> None:
        """Notify listeners when an album candidate has been matched."""
        plugins.send("album_matched", match=self)

    @property
    def item_info_pairs(self) -> list[tuple[Item, TrackInfo]]:
        """Return matched items together with their selected track metadata."""
        return list(self.mapping.items())

    @property
    def items(self) -> list[Item]:
        """Return the items that participate in this album match."""
        return [i for i, _ in self.item_info_pairs]

    @property
    def base_disambig_data(self) -> JSONDict:
        """Return album-specific values used in disambiguation displays."""
        return {
            "media": (
                f"{mediums}x{self.info.media}"
                if (mediums := self.info.mediums) and mediums > 1
                else self.info.media
            )
        }

    @property
    def merged_pairs(self) -> list[tuple[Item, JSONDict]]:
        """Generate item-data pairs with album-level fallback values."""
        return [
            (i, ti.merge_with_album(self.info))
            for i, ti in self.item_info_pairs
        ]

    def apply_metadata(self, from_scratch: bool | None = None) -> None:
        """Apply metadata to each of the items.

        If ``from_scratch`` is provided, its value determines whether the
        items existing metadata are cleared before applying new metadata.
        Otherwise, the configured ``from_scratch`` setting is used.
        """
        for item, data in self.merged_pairs:
            if self.from_scratch(from_scratch):
                item.clear()

            item.update(data)

    def apply_album_metadata(self, album: Album) -> None:
        """Apply album-level metadata to the Album object."""
        album.update(self.info.item_data)


@dataclass
class TrackMatch(Match):
    """Represent a singleton candidate and the item it updates."""

    disambig_fields_key = "singleton_disambig_fields"

    info: TrackInfo
    item: Item

    @property
    def base_disambig_data(self) -> JSONDict:
        """Return singleton-specific values used in disambiguation displays."""
        return {
            "index": f"Index {self.info.index}",
            "track_alt": f"Track {self.info.track_alt}",
            "album": (
                f"[{self.info.album}]"
                if (
                    config["import"]["singleton_album_disambig"].get()
                    and self.info.album
                )
                else ""
            ),
        }

    def apply_metadata(self, from_scratch: bool | None = None) -> None:
        """Apply metadata to the item.

        If ``from_scratch`` is provided, its value determines whether the
        item's existing metadata is cleared before applying new metadata.
        Otherwise, the configured ``from_scratch`` setting is used.
        """
        if self.from_scratch(from_scratch):
            self.item.clear()

        self.item.update(self.info.item_data)


# A structure for holding a set of possible matches to choose between. This
# consists of a list of possible candidates (i.e., AlbumInfo or TrackInfo
# objects) and a recommendation value.


class Proposal(NamedTuple):
    candidates: Sequence[AlbumMatch | TrackMatch]
    recommendation: Recommendation


def match_by_id(album_id: str | None, consensus: bool) -> Iterable[AlbumInfo]:
    """Return album candidates for the given album id.

    Make sure that the ID is present and that there is consensus on it among
    the items being tagged.
    """
    if not album_id:
        log.debug("No album ID found.")
    elif not consensus:
        log.debug("No album ID consensus.")
    else:
        log.debug("Searching for discovered album ID: {}", album_id)
        return metadata_plugins.albums_for_ids([album_id])

    return ()


def _recommendation(
    results: Sequence[AlbumMatch | TrackMatch],
) -> Recommendation:
    """Given a sorted list of AlbumMatch or TrackMatch objects, return a
    recommendation based on the results' distances.

    If the recommendation is higher than the configured maximum for
    an applied penalty, the recommendation will be downgraded to the
    configured maximum for that penalty.
    """
    if not results:
        # No candidates: no recommendation.
        return Recommendation.none

    # Basic distance thresholding.
    min_dist = results[0].distance
    if min_dist < config["match"]["strong_rec_thresh"].as_number():
        # Strong recommendation level.
        rec = Recommendation.strong
    elif min_dist <= config["match"]["medium_rec_thresh"].as_number():
        # Medium recommendation level.
        rec = Recommendation.medium
    elif len(results) == 1:
        # Only a single candidate.
        rec = Recommendation.low
    elif (
        results[1].distance - min_dist
        >= config["match"]["rec_gap_thresh"].as_number()
    ):
        # Gap between first two candidates is large.
        rec = Recommendation.low
    else:
        # No conclusion. Return immediately. Can't be downgraded any further.
        return Recommendation.none

    # Downgrade to the max rec if it is lower than the current rec for an
    # applied penalty.
    keys = set(min_dist.keys())
    if isinstance(results[0], AlbumMatch):
        for track_dist in min_dist.tracks.values():
            keys.update(list(track_dist.keys()))
    max_rec_view = config["match"]["max_rec"]
    for key in keys:
        if key in list(max_rec_view.keys()):
            max_rec = max_rec_view[key].as_choice(
                {
                    "strong": Recommendation.strong,
                    "medium": Recommendation.medium,
                    "low": Recommendation.low,
                    "none": Recommendation.none,
                }
            )
            rec = min(rec, max_rec)

    return rec


def _sort_candidates(candidates: Iterable[AnyMatch]) -> Sequence[AnyMatch]:
    """Sort candidates by distance."""
    return sorted(candidates, key=lambda match: match.distance)


def _add_candidate(
    items: Sequence[Item], results: Candidates[AlbumMatch], info: AlbumInfo
):
    """Given a candidate AlbumInfo object, attempt to add the candidate
    to the output dictionary of AlbumMatch objects. This involves
    checking the track count, ordering the items, checking for
    duplicates, and calculating the distance.
    """
    log.debug(
        "Candidate: {0.artist} - {0.album} ({0.album_id}) from {0.data_source}",
        info,
    )

    # Discard albums with zero tracks.
    if not info.tracks:
        log.debug("No tracks.")
        return

    # Prevent duplicates.
    if info.album_id and info.identifier in results:
        log.debug("Duplicate.")
        return

    # Discard matches without required tags.
    required_tags: Sequence[str] = config["match"]["required"].as_str_seq()
    for req_tag in required_tags:
        if getattr(info, req_tag) is None:
            log.debug("Ignored. Missing required tag: {}", req_tag)
            return

    # Find mapping between the items and the track info.
    item_info_pairs, extra_items, extra_tracks = assign_items(
        items, info.tracks
    )

    # Get the change distance.
    dist = distance(items, info, item_info_pairs)

    # Skip matches with ignored penalties.
    penalties = [key for key, _ in dist]
    ignored_tags: Sequence[str] = config["match"]["ignored"].as_str_seq()
    for penalty in ignored_tags:
        if penalty in penalties:
            log.debug("Ignored. Penalty: {}", penalty)
            return

    log.debug("Success. Distance: {}", dist)
    results[info.identifier] = AlbumMatch(
        dist, info, dict(item_info_pairs), extra_items, extra_tracks
    )


def tag_album(
    items,
    search_artist: str | None = None,
    search_name: str | None = None,
    search_ids: list[str] = [],
) -> tuple[str, str, Proposal]:
    """Return a tuple of the current artist name, the current album
    name, and a `Proposal` containing `AlbumMatch` candidates.

    The artist and album are the most common values of these fields
    among `items`.

    The `AlbumMatch` objects are generated by searching the metadata
    backends. By default, the metadata of the items is used for the
    search. This can be customized by setting the parameters.
    `search_ids` is a list of metadata backend IDs: if specified,
    it will restrict the candidates to those IDs, ignoring
    `search_artist` and `search album`. The `mapping` field of the
    album has the matched `items` as keys.

    The recommendation is calculated from the match quality of the
    candidates.
    """
    # Get current metadata.
    likelies, consensus = get_most_common_tags(items)
    cur_artist: str = likelies["artist"]
    cur_album: str = likelies["album"]
    log.debug("Tagging {} - {}", cur_artist, cur_album)

    # The output result, keys are (data_source, album_id) pairs, values are
    # AlbumMatch objects.
    candidates: Candidates[AlbumMatch] = {}

    # Search by explicit ID.
    if search_ids:
        log.debug("Searching for album IDs: {}", ", ".join(search_ids))
        for _info in metadata_plugins.albums_for_ids(search_ids):
            _add_candidate(items, candidates, _info)

    # Use existing metadata or text search.
    else:
        # Try search based on current ID.
        for info in match_by_id(
            likelies["mb_albumid"], consensus["mb_albumid"]
        ):
            _add_candidate(items, candidates, info)

        rec = _recommendation(list(candidates.values()))
        log.debug("Album ID match recommendation is {}", rec)
        if candidates and not config["import"]["timid"]:
            # If we have a very good MBID match, return immediately.
            # Otherwise, this match will compete against metadata-based
            # matches.
            if rec == Recommendation.strong:
                log.debug("ID match.")
                return (
                    cur_artist,
                    cur_album,
                    Proposal(list(candidates.values()), rec),
                )

        # Search terms.
        if not (search_artist and search_name):
            # No explicit search terms -- use current metadata.
            search_artist, search_name = cur_artist, cur_album
        log.debug("Search terms: {} - {}", search_artist, search_name)

        # Is this album likely to be a "various artist" release?
        va_likely = (
            (not consensus["artist"])
            or (search_artist.lower() in VA_ARTISTS)
            or any(item.comp for item in items)
        )
        log.debug("Album might be VA: {}", va_likely)

        # Get the results from the data sources.
        for matched_candidate in metadata_plugins.candidates(
            items, search_artist, search_name, va_likely
        ):
            _add_candidate(items, candidates, matched_candidate)

    log.debug("Evaluating {} candidates.", len(candidates))
    # Sort and get the recommendation.
    candidates_sorted = _sort_candidates(candidates.values())
    rec = _recommendation(candidates_sorted)
    return cur_artist, cur_album, Proposal(candidates_sorted, rec)


def tag_item(
    item,
    search_artist: str | None = None,
    search_name: str | None = None,
    search_ids: list[str] | None = None,
) -> Proposal:
    """Find metadata for a single track. Return a `Proposal` consisting
    of `TrackMatch` objects.

    `search_artist` and `search_title` may be used to override the item
    metadata in the search query. `search_ids` may be used for restricting the
    search to a list of metadata backend IDs.
    """
    # Holds candidates found so far: keys are (data_source, track_id) pairs,
    # values TrackMatch objects
    candidates: Candidates[TrackMatch] = {}
    rec: Recommendation | None = None

    # First, try matching by the external source ID.
    trackids = search_ids or [t for t in [item.mb_trackid] if t]
    if trackids:
        log.debug("Searching for track IDs: {}", ", ".join(trackids))
        for info in metadata_plugins.tracks_for_ids(trackids):
            dist = track_distance(item, info, incl_artist=True)
            candidates[info.identifier] = TrackMatch(dist, info, item)

        # If this is a good match, then don't keep searching.
        rec = _recommendation(_sort_candidates(candidates.values()))
        if rec == Recommendation.strong and not config["import"]["timid"]:
            log.debug("Track ID match.")
            return Proposal(_sort_candidates(candidates.values()), rec)

    # If we're searching by ID, don't proceed.
    if search_ids:
        if candidates:
            assert rec is not None
            return Proposal(_sort_candidates(candidates.values()), rec)
        return Proposal([], Recommendation.none)

    # Search terms.
    search_artist = search_artist or item.artist
    search_name = search_name or item.title
    log.debug("Item search terms: {} - {}", search_artist, search_name)

    # Get and evaluate candidate metadata.
    for track_info in metadata_plugins.item_candidates(
        item, search_artist, search_name
    ):
        dist = track_distance(item, track_info, incl_artist=True)
        candidates[track_info.identifier] = TrackMatch(dist, track_info, item)

    # Sort by distance and return with recommendation.
    log.debug("Found {} candidates.", len(candidates))
    candidates_sorted = _sort_candidates(candidates.values())
    rec = _recommendation(candidates_sorted)
    return Proposal(candidates_sorted, rec)



"""Tests for the 'subsonic' plugin."""

import unittest

import responses

from beets import config
from beetsplug import subsonicupdate


class SubsonicPluginTest(unittest.TestCase):
    """Test class for subsonicupdate."""

    @responses.activate
    def setUp(self):
        """Sets up config and plugin for test."""
        super().setUp()

        config["subsonic"]["user"] = "admin"
        config["subsonic"]["pass"] = "admin"
        config["subsonic"]["url"] = "http://localhost:4040"
        responses.add(
            responses.GET,
            "http://localhost:4040/rest/ping.view",
            status=200,
            body=self.PING_BODY,
        )
        self.subsonicupdate = subsonicupdate.SubsonicUpdate()

    PING_BODY = """
{
    "subsonic-response": {
        "status": "failed",
        "version": "1.15.0"
    }
}
"""
    SUCCESS_BODY = """
{
    "subsonic-response": {
        "status": "ok",
        "version": "1.15.0",
        "scanStatus": {
            "scanning": true,
            "count": 1000
        }
    }
}
"""

    FAILED_BODY = """
{
    "subsonic-response": {
        "status": "failed",
        "version": "1.15.0",
        "error": {
            "code": 40,
            "message": "Wrong username or password."
        }
    }
}
"""

    ERROR_BODY = """
{
    "timestamp": 1599185854498,
    "status": 404,
    "error": "Not Found",
    "message": "No message available",
    "path": "/rest/startScn"
}
"""

    @responses.activate
    def test_start_scan(self):
        """Tests success path based on best case scenario."""
        responses.add(
            responses.GET,
            "http://localhost:4040/rest/startScan",
            status=200,
            body=self.SUCCESS_BODY,
        )

        self.subsonicupdate.start_scan()

    @responses.activate
    def test_start_scan_failed_bad_credentials(self):
        """Tests failed path based on bad credentials."""
        responses.add(
            responses.GET,
            "http://localhost:4040/rest/startScan",
            status=200,
            body=self.FAILED_BODY,
        )

        self.subsonicupdate.start_scan()

    @responses.activate
    def test_start_scan_failed_not_found(self):
        """Tests failed path based on resource not found."""
        responses.add(
            responses.GET,
            "http://localhost:4040/rest/startScan",
            status=404,
            body=self.ERROR_BODY,
        )

        self.subsonicupdate.start_scan()

    def test_start_scan_failed_unreachable(self):
        """Tests failed path based on service not available."""
        self.subsonicupdate.start_scan()

    @responses.activate
    def test_url_with_context_path(self):
        """Tests success for included with contextPath."""
        config["subsonic"]["url"] = "http://localhost:4040/contextPath/"

        responses.add(
            responses.GET,
            "http://localhost:4040/contextPath/rest/startScan",
            status=200,
            body=self.SUCCESS_BODY,
        )

        self.subsonicupdate.start_scan()

    @responses.activate
    def test_url_with_trailing_forward_slash_url(self):
        """Tests success path based on trailing forward slash."""
        config["subsonic"]["url"] = "http://localhost:4040/"

        responses.add(
            responses.GET,
            "http://localhost:4040/rest/startScan",
            status=200,
            body=self.SUCCESS_BODY,
        )

        self.subsonicupdate.start_scan()

    @responses.activate
    def test_url_with_missing_port(self):
        """Tests failed path based on missing port."""
        config["subsonic"]["url"] = "http://localhost/airsonic"

        responses.add(
            responses.GET,
            "http://localhost/airsonic/rest/startScan",
            status=200,
            body=self.SUCCESS_BODY,
        )

        self.subsonicupdate.start_scan()

    @responses.activate
    def test_url_with_missing_schema(self):
        """Tests failed path based on missing schema."""
        config["subsonic"]["url"] = "localhost:4040/airsonic"

        responses.add(
            responses.GET,
            "http://localhost:4040/rest/startScan",
            status=200,
            body=self.SUCCESS_BODY,
        )

        self.subsonicupdate.start_scan()

    @responses.activate
    def test_start_scan_failed_non_json_response(self):
        """Tests failed path based on a non-JSON server response."""
        responses.add(
            responses.GET,
            "http://localhost:4040/rest/startScan",
            status=503,
            body="<html>server unavailable</html>",
            content_type="text/html",
        )

        with self.assertLogs("beets", level="ERROR") as logs:
            self.subsonicupdate.start_scan()

        assert "Subsonic server returned a non-JSON response" in "\n".join(
            logs.output
        )



"""Generates smart playlists based on beets queries."""

from __future__ import annotations

import os
from collections import defaultdict
from functools import cached_property
from shlex import quote as shell_quote
from typing import TYPE_CHECKING, Any, TypeAlias
from urllib.parse import quote
from urllib.request import pathname2url

import confuse

from beets import plugins, ui
from beets.dbcore.query import ParsingError, Query
from beets.dbcore.sort import Sort
from beets.exceptions import UserError
from beets.library import Album, Item, parse_query_string
from beets.util import (
    bytestring_path,
    mkdirall,
    normpath,
    path_as_posix,
    sanitize_path,
    syspath,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from beets.library import Library

QueryAndSort = tuple[Query, Sort]
PlaylistQuery = Query | tuple[QueryAndSort, ...] | None
PlaylistQueryAndSort = tuple[PlaylistQuery, Sort | None]
PlaylistMatch: TypeAlias = tuple[
    str, PlaylistQueryAndSort, PlaylistQueryAndSort
]


class SmartPlaylistPlugin(plugins.BeetsPlugin):
    def __init__(self) -> None:
        super().__init__()
        self.config.add(
            {
                "dest_regen": False,
                "relative_to": None,
                "playlist_dir": ".",
                "auto": True,
                "playlists": [],
                "uri_format": None,
                "fields": [],
                "forward_slash": False,
                "prefix": "",
                "urlencode": False,
                "format": "$artist - $title",
                "output": "m3u",
                "pretend": False,
            }
        )

        self.config["prefix"].redact = True  # May contain username/password.
        self._matched_playlists: set[PlaylistMatch] = set()
        self._unmatched_playlists: set[PlaylistMatch] = set()
        # validate output format
        self.config["output"].get(confuse.Choice(["m3u", "extm3u"]))

        if self.config["auto"]:
            self.register_listener("database_change", self.db_change)

    @cached_property
    def prefix(self) -> bytes:
        return bytestring_path(self.config["prefix"].as_str())

    @cached_property
    def relative_to(self) -> bytes | None:
        if relative_to := self.config["relative_to"].get():
            return normpath(relative_to)

        return None

    @cached_property
    def dest_regen(self) -> bool:
        return self.config["dest_regen"].get(bool)

    @cached_property
    def forward_slash(self) -> bool:
        return self.config["forward_slash"].get(bool)

    @cached_property
    def urlencode(self) -> bool:
        return self.config["urlencode"].get(bool)

    @cached_property
    def uri_format(self) -> str | None:
        return self.config["uri_format"].get()

    def commands(self) -> list[ui.Subcommand]:
        spl_update = ui.Subcommand(
            "splupdate",
            help="update the smart playlists. Playlist names may be "
            "passed as arguments.",
        )
        spl_update.parser.add_option(
            "-p",
            "--pretend",
            action="store_true",
            help="display query results but don't write playlist files.",
        )
        spl_update.parser.add_option(
            "-f",
            "--format",
            type="string",
            default=self.config["format"].get(),
            help="print per-track log lines with custom format",
        )
        spl_update.parser.add_option(
            "-d",
            "--playlist-dir",
            dest="playlist_dir",
            metavar="PATH",
            type="string",
            default=self.config["playlist_dir"].get(),
            help="directory to write the generated playlist files to.",
        )
        spl_update.parser.add_option(
            "--dest-regen",
            action="store_true",
            dest="dest_regen",
            default=self.config["dest_regen"].get(bool),
            help="regenerate the destination path as 'move' or 'convert' "
            "commands would do.",
        )
        spl_update.parser.add_option(
            "--relative-to",
            dest="relative_to",
            metavar="PATH",
            type="string",
            default=self.config["relative_to"].get(),
            help="generate playlist item paths relative to this path.",
        )
        spl_update.parser.add_option(
            "--prefix",
            type="string",
            default=self.config["prefix"].get(),
            help="prepend string to every path in the playlist file.",
        )
        spl_update.parser.add_option(
            "--forward-slash",
            action="store_true",
            dest="forward_slash",
            default=self.config["forward_slash"].get(bool),
            help="force forward slash in paths within playlists.",
        )
        spl_update.parser.add_option(
            "--urlencode",
            action="store_true",
            default=self.config["urlencode"].get(bool),
            help="URL-encode all paths.",
        )
        spl_update.parser.add_option(
            "--uri-format",
            dest="uri_format",
            type="string",
            default=self.config["uri_format"].get(),
            help="playlist item URI template, e.g. http://beets:8337/item/$id/file.",
        )
        spl_update.parser.add_option(
            "--output",
            type="choice",
            choices=["m3u", "extm3u"],
            default=self.config["output"].get(),
            help="specify the playlist format: m3u|extm3u.",
        )
        spl_update.func = self.update_cmd
        return [spl_update]

    def update_cmd(self, lib: Library, opts: Any, args: list[str]) -> None:
        self.build_queries()
        if args:
            args_set = set(args)
            for a in list(args_set):
                if not a.endswith(".m3u"):
                    args_set.add(f"{a}.m3u")

            playlists = {
                (name, q, a_q)
                for name, q, a_q in self._unmatched_playlists
                if name in args_set
            }
            if not playlists:
                unmatched = [name for name, _, _ in self._unmatched_playlists]
                unmatched.sort()
                quoted_names = " ".join(shell_quote(name) for name in unmatched)
                raise UserError(
                    f"No playlist matching any of {quoted_names} found"
                )

            self._matched_playlists = playlists
            self._unmatched_playlists -= playlists
        else:
            self._matched_playlists = self._unmatched_playlists

        self.config.set(vars(opts))
        self.update_playlists(lib)

    def _parse_one_query(
        self, playlist: dict[str, Any], key: str, model_cls: type
    ) -> tuple[PlaylistQuery, Sort | None]:
        qs = playlist.get(key)
        if qs is None:
            return None, None
        if isinstance(qs, str):
            return parse_query_string(qs, model_cls)
        if len(qs) == 1:
            return parse_query_string(qs[0], model_cls)

        queries_and_sorts: tuple[QueryAndSort, ...] = tuple(
            parse_query_string(q, model_cls) for q in qs
        )
        return queries_and_sorts, None

    def build_queries(self) -> None:
        """
        Instantiate queries for the playlists.

        Each playlist has 2 queries: one for items, one for albums, each with a
        sort. We must also remember its name. _unmatched_playlists is a set of
        tuples (name, (q, q_sort), (album_q, album_q_sort)).

        sort may be any sort, or NullSort, or None. None and NullSort are
        equivalent and both eval to False.
        More precisely
        - it will be NullSort when a playlist query ('query' or 'album_query')
          is a single item or a list with 1 element
        - it will be None when there are multiple items in a query
        """
        self._unmatched_playlists = set()
        self._matched_playlists = set()

        for playlist in self.config["playlists"].get(list):
            if "name" not in playlist:
                self._log.warning("playlist configuration is missing name")
                continue

            try:
                q_match = self._parse_one_query(playlist, "query", Item)
                a_match = self._parse_one_query(playlist, "album_query", Album)
            except ParsingError as exc:
                self._log.warning(
                    "invalid query in playlist {}: {}", playlist["name"], exc
                )
                continue

            self._unmatched_playlists.add((playlist["name"], q_match, a_match))

    def _matches_query(self, model: Item | Album, query: PlaylistQuery) -> bool:
        if not query:
            return False
        if isinstance(query, (list, tuple)):
            return any(q.match(model) for q, _ in query)
        return query.match(model)

    def matches(
        self,
        model: Item | Album,
        query: PlaylistQuery,
        album_query: PlaylistQuery,
    ) -> bool:
        if isinstance(model, Album):
            return self._matches_query(model, album_query)
        if isinstance(model, Item):
            return self._matches_query(model, query)
        return False

    def db_change(self, lib: Library, model: Item | Album) -> None:
        if self._unmatched_playlists is None:
            self.build_queries()

        for playlist in self._unmatched_playlists:
            n, (q, _), (a_q, _) = playlist
            if self.matches(model, q, a_q):
                self._log.debug("{} will be updated because of {}", n, model)
                self._matched_playlists.add(playlist)
                self.register_listener("cli_exit", self.update_playlists)

        self._unmatched_playlists -= self._matched_playlists

    @staticmethod
    def get_queries(
        query: PlaylistQueryAndSort,
    ) -> list[tuple[Query, Sort | None]]:
        """Normalize a playlist query into a flat list of query-sort pairs.

        Handles both compound (list/tuple of pairs) and single query inputs,
        returning an empty list when no query is present.
        """
        q, sort = query

        if isinstance(q, (list, tuple)):
            return list(q)
        if q:
            return [(q, sort)]

        return []

    @classmethod
    def get_playlist_items(
        cls,
        lib: Library,
        item_q: PlaylistQueryAndSort,
        album_q: PlaylistQueryAndSort,
    ) -> Iterator[Item]:
        """Collect unique items matching the playlist's item and album queries.

        Queries both tracks directly and albums (expanding them to their
        tracks), then merges the results into a deduplicated list preserving
        insertion order.
        """
        items: list[Item] = []

        for q, sort in cls.get_queries(item_q):
            items.extend(lib.items(q, sort))

        albums: list[Album] = []
        for q, sort in cls.get_queries(album_q):
            albums.extend(lib.albums(q, sort))

        seen_album_ids = set()
        for album in albums:
            if album.id not in seen_album_ids:
                seen_album_ids.add(album.id)
                items.extend(album.items())

        seen_ids = set()
        for item in items:
            if item.id not in seen_ids:
                seen_ids.add(item.id)
                yield item

    def get_item_uri(self, item: Item) -> bytes:
        item_uri = item.path
        if uri_format := self.uri_format:
            return uri_format.replace("$id", str(item.id)).encode("utf-8")

        if self.dest_regen:
            item_uri = item.destination()
        if self.relative_to:
            item_uri = os.path.relpath(item_uri, self.relative_to)
        if self.forward_slash:
            item_uri = path_as_posix(item_uri)
        if self.urlencode:
            item_uri = bytestring_path(pathname2url(os.fsdecode(item_uri)))

        return self.prefix + item_uri

    def write_playlist(
        self, path: bytes, is_extm3u: bool, entries: list[PlaylistItem]
    ) -> None:
        """Write a playlist file with the given entries."""
        mkdirall(path)
        with open(syspath(path), "wb") as f:
            keys = []
            if is_extm3u:
                keys = self.config["fields"].get(list)
                f.write(b"#EXTM3U\n")
            for entry in entries:
                f.write(entry.get_comment(is_extm3u, keys))

    def update_playlists(self, lib: Library) -> None:
        playlist_count = len(self._matched_playlists)
        self._log.info("Updating {} smart playlists...", playlist_count)

        playlist_dir = bytestring_path(
            self.config["playlist_dir"].as_filename()
        )

        # Maps playlist filenames to lists of track entries and URI sets used
        # to deduplicate output lines.
        m3us: dict[str, list[PlaylistItem]] = defaultdict(list)
        m3u_uris_by_name: dict[str, set[bytes]] = defaultdict(set)

        for playlist in self._matched_playlists:
            name, item_q, album_q = playlist
            items = self.get_playlist_items(lib, item_q, album_q)

            # As we allow tags in the m3u names, we'll need to iterate through
            # the items and generate the correct m3u file names.
            matched_items: list[Item] = []
            for item in items:
                m3u_name = item.evaluate_template(name, True)
                m3u_name = sanitize_path(m3u_name, lib.replacements)
                item_uri = self.get_item_uri(item)

                if item_uri not in m3u_uris_by_name[m3u_name]:
                    m3u_uris_by_name[m3u_name].add(item_uri)
                    m3us[m3u_name].append(PlaylistItem(item, item_uri))
                    matched_items.append(item)

            self._log.info(
                "Creating playlist {}: {} tracks.", name, len(matched_items)
            )
            for item in matched_items:
                self._log.debug(
                    item.evaluate_template(self.config["format"].as_str())
                )

        if self.config["pretend"].get():
            self._log.info("{} playlists would be updated", playlist_count)
        else:
            # Write all of the accumulated track lists to files.
            is_extm3u = self.config["output"].get() == "extm3u"
            for m3u, entries in m3us.items():
                m3u_path = normpath(
                    os.path.join(playlist_dir, bytestring_path(m3u))
                )
                self.write_playlist(m3u_path, is_extm3u, entries)

            # Send an event when playlists were updated.
            plugins.send("smartplaylist_update")
            self._log.info("{} playlists updated", playlist_count)


class PlaylistItem:
    def __init__(self, item: Item, uri: bytes) -> None:
        self.item = item
        self.uri = uri

    def get_comment(self, is_extm3u: bool, fields: list[str]) -> bytes:
        comment = ""
        if is_extm3u:
            attr = [(k, self.item[k]) for k in fields]
            al = [
                f' {k}="{quote("; ".join(v) if isinstance(v, list) else str(v), safe="/:")}"'  # noqa: E501
                for k, v in attr
            ]
            attrs = "".join(al)
            comment = (
                f"#EXTINF:{int(self.item.length)}{attrs},"
                f"{self.item.artist} - {self.item.title}\n"
            )

        return comment.encode("utf-8") + self.uri + b"\n"



"""Updates Subsonic library on Beets import
Your Beets configuration file should contain
a "subsonic" section like the following:
    subsonic:
        url: https://mydomain.com:443/subsonic
        user: username
        pass: password
        auth: token
For older Subsonic versions, token authentication
is not supported, use password instead:
    subsonic:
        url: https://mydomain.com:443/subsonic
        user: username
        pass: password
        auth: pass
"""

import hashlib
import random
import string
from binascii import hexlify

import requests

from beets.plugins import BeetsPlugin

__author__ = "https://github.com/maffo999"


class SubsonicUpdate(BeetsPlugin):
    def __init__(self):
        super().__init__("subsonic")
        # Set default configuration values
        self.config.add(
            {
                "user": "admin",
                "pass": "admin",
                "url": "http://localhost:4040",
                "auth": "token",
            }
        )
        self.config["user"].redact = True
        self.config["pass"].redact = True
        self.register_listener("database_change", self.db_change)
        self.register_listener("smartplaylist_update", self.spl_update)

    def db_change(self, lib, model):
        self.register_listener("cli_exit", self.start_scan)

    def spl_update(self):
        self.register_listener("cli_exit", self.start_scan)

    def __create_token(self):
        """Create salt and token from given password.

        :return: The generated salt and hashed token
        """
        password = self.config["pass"].as_str()

        # Pick the random sequence and salt the password
        r = string.ascii_letters + string.digits
        salt = "".join([random.choice(r) for _ in range(6)])
        salted_password = f"{password}{salt}"
        token = hashlib.md5(salted_password.encode("utf-8")).hexdigest()

        # Put together the payload of the request to the server and the URL
        return salt, token

    def __format_url(self, endpoint):
        """Get the Subsonic URL to trigger the given endpoint.
        Uses either the url config option or the deprecated host, port,
        and context_path config options together.

        :return: Endpoint for updating Subsonic
        """

        url = self.config["url"].as_str()
        if url and url.endswith("/"):
            url = url[:-1]

        # @deprecated("Use url config option instead")
        if not url:
            host = self.config["host"].as_str()
            port = self.config["port"].get(int)
            context_path = self.config["contextpath"].as_str()
            if context_path == "/":
                context_path = ""
            url = f"http://{host}:{port}{context_path}"

        return f"{url}/rest/{endpoint}"

    def start_scan(self):
        user = self.config["user"].as_str()
        auth = self.config["auth"].as_str()
        url = self.__format_url("startScan")
        self._log.debug("URL is {}", url)
        self._log.debug("auth type is {.config[auth]}", self)

        if auth == "token":
            salt, token = self.__create_token()
            payload = {
                "u": user,
                "t": token,
                "s": salt,
                "v": "1.13.0",  # Subsonic 5.3 and newer
                "c": "beets",
                "f": "json",
            }
        elif auth == "password":
            password = self.config["pass"].as_str()
            encpass = hexlify(password.encode()).decode()
            payload = {
                "u": user,
                "p": f"enc:{encpass}",
                "v": "1.12.0",
                "c": "beets",
                "f": "json",
            }
        else:
            return
        try:
            response = requests.get(url, params=payload, timeout=10)
            json = response.json()
        except requests.exceptions.JSONDecodeError:
            self._log.error(
                "Subsonic server returned a non-JSON response from {} "
                "(HTTP {})",
                url,
                response.status_code,
            )
            return
        except requests.exceptions.RequestException as error:
            self._log.error("Error connecting to Subsonic server: {}", error)
            return
        try:
            if (
                response.status_code == 200
                and json["subsonic-response"]["status"] == "ok"
            ):
                count = json["subsonic-response"]["scanStatus"]["count"]
                self._log.info("Updating Subsonic; scanning {} tracks", count)
            elif (
                response.status_code == 200
                and json["subsonic-response"]["status"] == "failed"
            ):
                self._log.error(
                    "Error: {[subsonic-response][error][message]}", json
                )
            else:
                self._log.error("Error: {}", json)
        except Exception as error:
            self._log.error("Error: {}", error)
